import typing as t
import warnings
from importlib import import_module

from paradigm._core import (catalog,
                            namespacing,
                            pretty)

QualifiedPaths = t.Dict[
    catalog.Path, t.Dict[catalog.Path, t.List[catalog.QualifiedPath]]
]


def from_modules(modules_paths: t.Iterable[catalog.Path]) -> QualifiedPaths:
    result: QualifiedPaths = {}
    for module_path in modules_paths:
        _index_module_path(module_path,
                           paths=result)
    return result


def _index_module_or_type(namespace: namespacing.ModuleOrType,
                          *,
                          paths: QualifiedPaths,
                          module_path: catalog.Path,
                          parent_path: catalog.Path,
                          visited_classes: t.Set[type]) -> None:
    for name, value in vars(namespace).items():
        if isinstance(value, type):
            if value not in visited_classes:
                object_path = parent_path + (name,)
                qualified_module_path, qualified_object_path = (
                    catalog.qualified_path_from(value)
                )
                assert (
                        qualified_module_path or qualified_object_path
                ), catalog.path_to_string(module_path + object_path)
                (paths.setdefault(qualified_module_path, {})
                 .setdefault(qualified_object_path, [])
                 .append((module_path, object_path)))
                _index_module_or_type(
                        value,
                        paths=paths,
                        module_path=module_path,
                        parent_path=object_path,
                        visited_classes={*visited_classes, value}
                )
        else:
            qualified_module_path, qualified_object_path = (
                catalog.qualified_path_from(value)
            )
            if qualified_object_path:
                (paths.setdefault(qualified_module_path, {})
                 .setdefault(qualified_object_path, [])
                 .append((module_path, parent_path + (name,))))


def _index_module_path(module_path: catalog.Path,
                       *,
                       paths: QualifiedPaths) -> None:
    module_name = catalog.path_to_string(module_path)
    try:
        module = import_module(module_name)
    except Exception as error:
        warnings.warn(f'Failed importing module "{module_name}". '
                      f'Reason:\n{pretty.format_exception(error)}',
                      ImportWarning)
    else:
        _index_module_or_type(module,
                              paths=paths,
                              module_path=module_path,
                              parent_path=(),
                              visited_classes=set())



from __future__ import annotations

import functools
import sys
import typing as t

import typing_extensions as te

_Params = te.ParamSpec('_Params')
_T1 = t.TypeVar('_T1')
_T2 = t.TypeVar('_T2')


def decorate_if(
        decorator: t.Callable[[t.Callable[_Params, _T1]], t.Any],
        condition: bool
) -> t.Callable[[t.Callable[_Params, _T1]], t.Any]:
    return decorator if condition else _identity_decorator


singledispatchmethod: t.Any
if sys.version_info < (3, 8):

    class _singledispatchmethod:
        dispatcher: t.Any
        func: t.Callable[..., _T1]

        def __new__(cls, func: t.Callable[..., t.Any]) -> te.Self:
            if not callable(func) and not hasattr(func, '__get__'):
                raise TypeError(f'{func!r} is not callable or a descriptor')
            self = super().__new__(cls)
            self.dispatcher, self.func = functools.singledispatch(func), func
            return self

        @t.overload
        def register(self,
                     cls: t.Type[t.Any],
                     method: t.Callable[..., _T1]) -> t.Any:
            return self.dispatcher.register(cls, method)

        @t.overload
        def register(self, cls: t.Type[t.Any]) -> t.Any:
            return self.dispatcher.register(cls)

        def register(self,
                     cls: t.Type[t.Any],
                     method: t.Optional[t.Callable[..., _T1]] = None) -> t.Any:
            return self.dispatcher.register(cls, method)

        @property
        def __isabstractmethod__(self) -> bool:
            return getattr(self.func, '__isabstractmethod__', False)

        def __get__(self,
                    instance: _T2,
                    cls: t.Optional[_T2] = None) -> t.Any:
            def dispatchable_method(*args: t.Any,
                                    **kwargs: t.Any) -> t.Any:
                method = self.dispatcher.dispatch(args[0].__class__)
                return method.__get__(instance, cls)(*args, **kwargs)

            result: t.Any = dispatchable_method
            result.__isabstractmethod__ = self.__isabstractmethod__
            result.register = self.register
            functools.update_wrapper(result, self.func)
            return result


    singledispatchmethod = _singledispatchmethod
else:
    singledispatchmethod = functools.singledispatchmethod


def _identity_decorator(
        value: t.Callable[_Params, _T1]
) -> t.Callable[_Params, _T1]:
    return value



import ast
import sys
import typing as t
from collections import deque

import typing_extensions as te

from paradigm._core import (catalog,
                            sources)


def recursively_iterate_children(node: ast.AST) -> t.Iterable[ast.AST]:
    candidates = deque(ast.iter_child_nodes(node))
    while candidates:
        candidate = candidates.popleft()
        yield candidate
        candidates.extend(ast.iter_child_nodes(candidate))


def to_parent_module_path(
        ast_node: ast.ImportFrom,
        *,
        parent_module_path: catalog.Path
) -> catalog.Path:
    level = ast_node.level
    import_is_relative = level > 0
    if not import_is_relative:
        assert ast_node.module is not None, ast_node
        return catalog.path_from_string(ast_node.module)
    depth = (len(parent_module_path)
             + sources.is_package(parent_module_path)
             - level) or None
    module_path_parts = (list(parent_module_path[:depth])
                         + ([]
                            if ast_node.module is None
                            else ast_node.module.split('.')))
    return tuple(module_path_parts)


if sys.version_info < (3, 9):
    def subscript_to_item(ast_node: ast.Subscript) -> ast.expr:
        assert isinstance(ast_node.slice, ast.Index)
        return ast_node.slice.value
else:
    def subscript_to_item(ast_node: ast.Subscript) -> ast.expr:
        return ast_node.slice


def is_dependency_name(ast_node: ast.AST) -> te.TypeGuard[ast.Name]:
    return (isinstance(ast_node, ast.Name)
            and isinstance(ast_node.ctx, ast.Load))



import ast
import builtins
import operator
import sys
import types
import typing as t
import weakref
from collections import ChainMap
from functools import singledispatch
from importlib import import_module

from typing_extensions import Self

from paradigm._core import (catalog,
                            namespacing,
                            scoping,
                            sources,
                            stubs)
from . import (construction,
               conversion)
from .execution import execute_statement
from .kind import NodeKind
from .utils import (is_dependency_name,
                    recursively_iterate_children)

if sys.version_info < (3, 9):
    AstExpression = t.Union[ast.expr, ast.slice]
else:
    AstExpression = ast.expr


@singledispatch
def evaluate_expression_node(ast_node: AstExpression,
                             module_path: catalog.Path,
                             parent_path: catalog.Path,
                             parent_namespace: namespacing.Namespace) -> t.Any:
    raise TypeError(type(ast_node))


@singledispatch
def evaluate_statement_node(ast_node: ast.stmt,
                            module_path: catalog.Path,
                            parent_path: catalog.Path,
                            parent_namespace: namespacing.Namespace) -> t.Any:
    raise TypeError(type(ast_node))


@singledispatch
def evaluate_operator_node(
        ast_node: t.Union[ast.operator, ast.unaryop]
) -> t.Any:
    raise TypeError(type(ast_node))


@evaluate_operator_node.register(ast.USub)
def _(ast_node: ast.USub) -> t.Any:
    return operator.neg


@evaluate_operator_node.register(ast.UAdd)
def _(ast_node: ast.UAdd) -> t.Any:
    return operator.pos


@evaluate_expression_node.register(ast.UnaryOp)
def _(ast_node: ast.UnaryOp,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    return evaluate_operator_node(ast_node.op)(
            evaluate_expression_node(ast_node.operand, module_path,
                                     parent_path, parent_namespace)
    )


@evaluate_operator_node.register(ast.BitAnd)
def _(ast_node: ast.BitAnd) -> t.Any:
    return operator.and_


@evaluate_operator_node.register(ast.BitOr)
def _(ast_node: ast.BitOr) -> t.Any:
    return operator.or_


@evaluate_operator_node.register(ast.BitXor)
def _(ast_node: ast.BitXor) -> t.Any:
    return operator.xor


@evaluate_expression_node.register(ast.Call)
def _(ast_node: ast.Call,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    args = [evaluate_expression_node(arg, module_path, parent_path,
                                     parent_namespace)
            for arg in ast_node.args]
    kwargs = {
        keyword.arg: evaluate_expression_node(keyword.value, module_path,
                                              parent_path, parent_namespace)
        for keyword in ast_node.keywords
    }
    return evaluate_expression_node(
            ast_node.func, module_path, parent_path, parent_namespace
    )(*args, **kwargs)


class _LazyEvaluator(ast.NodeTransformer):
    def visit_AsyncFunctionDef(
            self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:
        if node.returns is not None:
            node.returns = ast.Str(conversion.to_str(node.returns))
        self.generic_visit(node)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        if node.returns is not None:
            node.returns = ast.Str(conversion.to_str(node.returns))
        self.generic_visit(node)
        return node

    if sys.version_info < (3, 8):
        def visit_arg(self, node: ast.arg) -> ast.arg:
            return ast.arg(node.arg,
                           None
                           if node.annotation is None
                           else ast.Str(conversion.to_str(node.annotation)))
    else:
        def visit_arg(self, node: ast.arg) -> ast.arg:
            return ast.arg(node.arg,
                           None
                           if node.annotation is None
                           else ast.Str(conversion.to_str(node.annotation)),
                           getattr(node, 'type_comment', None))


_AstNode = t.TypeVar('_AstNode',
                     bound=ast.AST)
_to_lazy_statement: t.Callable[[_AstNode], _AstNode] = _LazyEvaluator().visit


@evaluate_statement_node.register(ast.AsyncFunctionDef)
@evaluate_statement_node.register(ast.FunctionDef)
def _(ast_node: t.Union[ast.AsyncFunctionDef, ast.FunctionDef],
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    namespace = dict(parent_namespace)
    try:
        source_path = sources.from_module_path(module_path)
    except sources.NotFound:
        source_path = sources.Path(*module_path)
    execute_statement(ast.fix_missing_locations(_to_lazy_statement(ast_node)),
                      source_path=source_path,
                      namespace=namespace)
    name = ast_node.name
    result = parent_namespace[name] = namespace.pop(name)
    return result


@evaluate_expression_node.register(ast.Constant)
def _(ast_node: ast.Constant,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    return ast_node.value


@evaluate_expression_node.register(ast.Tuple)
def _(ast_node: ast.Tuple,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return tuple(evaluate_expression_node(element, module_path, parent_path,
                                          parent_namespace)
                 for element in ast_node.elts)


@evaluate_expression_node.register(ast.List)
def _(ast_node: ast.List,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return [evaluate_expression_node(element, module_path, parent_path,
                                     parent_namespace)
            for element in ast_node.elts]


@evaluate_statement_node.register(ast.AnnAssign)
@evaluate_statement_node.register(ast.Assign)
def _(ast_node: ast.Assign,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace,
      *,
      cache: t.Dict[catalog.QualifiedPath, t.Any] = {}) -> t.Any:
    targets_paths = [
        scoping.resolve_object_path(module_path, parent_path, (name,),
                                    stubs.definitions, stubs.references,
                                    stubs.submodules, stubs.superclasses)
        for name in conversion.to_names(ast_node)
    ]
    for target_path in targets_paths:
        try:
            result = cache[target_path]
        except KeyError:
            _, target_object_path = target_path
            cache[target_path] = t.ForwardRef(
                    catalog.path_to_string(target_object_path)
            )
            continue
        else:
            break
    else:
        result = evaluate_expression_node(ast_node.value, module_path,
                                          parent_path, parent_namespace)
    for target_path in targets_paths:
        cache[target_path] = result
    return result


@evaluate_expression_node.register(ast.BinOp)
def _(ast_node: ast.BinOp,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    left = evaluate_expression_node(ast_node.left, module_path, parent_path,
                                    parent_namespace)
    right = evaluate_expression_node(ast_node.right, module_path, parent_path,
                                     parent_namespace)
    try:
        return evaluate_operator_node(ast_node.op)(left, right)
    except TypeError:
        assert isinstance(ast_node.op, ast.BitOr), ast_node
        return t.Union[left, right]


@evaluate_expression_node.register(ast.Name)
def _(ast_node: ast.Name,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    object_name = ast_node.id
    module_path, object_path = scoping.resolve_object_path(
            module_path, parent_path, (object_name,), stubs.definitions,
            stubs.references, stubs.submodules, stubs.superclasses
    )
    return evaluate_qualified_path(module_path, object_path, parent_namespace)


@evaluate_expression_node.register(ast.Attribute)
def _(ast_node: ast.Attribute,
      module_path: catalog.Path,
      parent_path: catalog.Path,
      parent_namespace: namespacing.Namespace) -> t.Any:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    value = evaluate_expression_node(ast_node.value, module_path, parent_path,
                                     parent_namespace)
    try:
        return getattr(value, ast_node.attr)
    except AttributeError:
        module_path, object_path = scoping.resolve_object_path(
                module_path, parent_path, conversion.to_path(ast_node),
                stubs.definitions, stubs.references, stubs.submodules,
                stubs.superclasses
        )
        return evaluate_qualified_path(module_path, object_path,
                                       parent_namespace)


def evaluate_qualified_path(
        module_path: catalog.Path,
        object_path: catalog.Path,
        parent_namespace: namespacing.Namespace,
        *,
        builtins_module_path: catalog.Path
        = catalog.module_path_from_module(builtins),
        modules_cache: t.MutableMapping[catalog.Path, types.ModuleType]
        = weakref.WeakValueDictionary(),
        objects_cache: t.MutableMapping[catalog.QualifiedPath, t.Any]
        = weakref.WeakValueDictionary([((('_typeshed',), ('Self',)), Self)])
) -> t.Any:
    module_name = catalog.path_to_string(module_path)
    try:
        module = import_module(module_name)
    except ImportError:
        if not object_path:
            try:
                return modules_cache[module_path]
            except KeyError:
                modules_cache[module_path] = module = types.ModuleType(
                        module_name
                )
                namespace = module.__dict__
                for name in stubs.definitions[module_path].keys():
                    namespace[name] = parent_namespace[name] = (
                        evaluate_qualified_path(module_path, (name,),
                                                parent_namespace)
                    )
                return module
    else:
        try:
            return (namespacing.search(module, object_path)
                    if object_path
                    else module)
        except namespacing.ObjectNotFound:
            pass
    try:
        return objects_cache[(module_path, object_path)]
    except KeyError:
        module_raw_ast_nodes = stubs.raw_ast_nodes[module_path]
        node_kind = stubs.nodes_kinds[module_path][object_path]
        try:
            raw_ast_nodes = module_raw_ast_nodes[object_path]
        except KeyError:
            assert node_kind is NodeKind.CLASS, (module_path, object_path)
        else:
            if node_kind is not NodeKind.CLASS:
                for raw_ast_node in raw_ast_nodes:
                    ast_node = construction.from_raw(raw_ast_node)
                    assert isinstance(ast_node, ast.stmt), (module_path,
                                                            object_path)
                    value = parent_namespace[object_path[-1]] = (
                        evaluate_statement_node(ast_node, module_path,
                                                object_path[:-1],
                                                parent_namespace)
                    )
                return value
        scope = stubs.definitions[module_path]
        for part in object_path:
            scope = scope[part]
        class_namespace: namespacing.Namespace = {'__module__': __name__}
        module_namespace: namespacing.Namespace = {'__name__': __name__}
        annotations_nodes = {}
        for name in scope.keys():
            raw_ast_nodes = module_raw_ast_nodes[(*object_path, name)]
            ast_nodes = [construction.from_raw(raw_ast_node)
                         for raw_ast_node in raw_ast_nodes]
            definitions_nodes: t.List[ast.stmt] = []
            for ast_node in ast_nodes:
                if (isinstance(ast_node, ast.AnnAssign)
                        and ast_node.value is None):
                    annotations_nodes[name] = ast_node.annotation
                else:
                    assert isinstance(ast_node, ast.stmt), (module_path,
                                                            object_path)
                    definitions_nodes.append(ast_node)
            for definition_node in definitions_nodes:
                for dependency_name in {
                    child.id
                    for child in recursively_iterate_children(
                            _to_lazy_statement(definition_node)
                    )
                    if is_dependency_name(child)
                }:
                    dependency_module_path, dependency_object_path = (
                        scoping.resolve_object_path(
                                module_path, object_path,
                                (dependency_name,), stubs.definitions,
                                stubs.references, stubs.submodules,
                                stubs.superclasses
                        )
                    )
                    if dependency_module_path != builtins_module_path:
                        module_namespace[dependency_name] = (
                            evaluate_qualified_path(
                                    dependency_module_path,
                                    dependency_object_path,
                                    module_namespace
                            )
                        )
                value = evaluate_statement_node(
                        definition_node, module_path, object_path,
                        ChainMap(class_namespace, module_namespace)
                )
            if definitions_nodes:
                class_namespace[name] = value
        bases = tuple(
                evaluate_qualified_path(superclass_module_path,
                                        superclass_object_path,
                                        module_namespace)
                for (
                    superclass_module_path, superclass_object_path
                ) in stubs.superclasses.get(module_path, {}).get(
                        object_path, []
                )
        )
        if annotations_nodes:
            class_namespace['__annotations__'] = {}
        result = parent_namespace[object_path[-1]] = objects_cache[
            (module_path, object_path)
        ] = types.new_class(
                catalog.path_to_string(object_path), bases,
                exec_body=lambda namespace: namespace.update(
                        class_namespace
                )
        )
        if annotations_nodes:
            result.__annotations__.update({
                name: evaluate_expression_node(annotation_node,
                                               module_path, object_path,
                                               parent_namespace)
                for name, annotation_node in annotations_nodes.items()
            })
        globals()[catalog.path_to_string(object_path)] = result
        return result


if sys.version_info < (3, 8):
    @evaluate_expression_node.register(ast.Ellipsis)
    def _(ast_node: ast.Ellipsis,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        return Ellipsis


    @evaluate_expression_node.register(ast.NameConstant)
    def _(ast_node: ast.NameConstant,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        return ast_node.value


    @evaluate_expression_node.register(ast.Num)
    def _(ast_node: ast.Num,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        return ast_node.n


    @evaluate_expression_node.register(ast.Str)
    def _(ast_node: ast.Str,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        return ast_node.s

if sys.version_info < (3, 9):
    _GenericAlias: t.Any = type(t.List)
    _types_to_generic_aliases = {value.__origin__: value
                                 for value in vars(t).values()
                                 if isinstance(value, _GenericAlias)}


    @evaluate_expression_node.register(ast.Index)
    def _(ast_node: ast.Index,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        return evaluate_expression_node(ast_node.value, module_path,
                                        parent_path, parent_namespace)


    @evaluate_expression_node.register(ast.Subscript)
    def _(ast_node: ast.Subscript,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        assert isinstance(ast_node.ctx, ast.Load), ast_node
        value = evaluate_expression_node(ast_node.value, module_path,
                                         parent_path, parent_namespace)
        item = evaluate_expression_node(ast_node.slice, module_path,
                                        parent_path, parent_namespace)
        try:
            generic_alias = _types_to_generic_aliases[value]
        except KeyError:
            try:
                return value[item]
            except TypeError:
                return value
        else:
            return generic_alias[item]
else:
    @evaluate_expression_node.register(ast.Subscript)
    def _(ast_node: ast.Subscript,
          module_path: catalog.Path,
          parent_path: catalog.Path,
          parent_namespace: namespacing.Namespace) -> t.Any:
        assert isinstance(ast_node.ctx, ast.Load), ast_node
        value = evaluate_expression_node(ast_node.value, module_path,
                                         parent_path, parent_namespace)
        item = evaluate_expression_node(ast_node.slice, module_path,
                                        parent_path, parent_namespace)
        try:
            return value[item]
        except TypeError:
            return value



import typing as t
import warnings
from compileall import compile_file
from importlib import import_module
from importlib.machinery import SOURCE_SUFFIXES
from operator import attrgetter
from pathlib import Path

import typing_extensions as te

from . import pretty

FILE_SUFFIX: te.Final[str] = SOURCE_SUFFIXES[0]


def load(*names: str, path: Path) -> t.Tuple[t.Any, ...]:
    return attrgetter(*names)(
            import_module((''
                           if __name__ in ('__main__', '__mp_main__')
                           else __name__.rsplit('.', maxsplit=1)[0] + '.')
                          + path.stem)
    )


def save(*, path: Path, **values: t.Any) -> None:
    try:
        with path.open('w',
                       encoding='utf-8') as file:
            for name, value in values.items():
                file.write(f'{name} = '
                           + pretty.repr_from(value, 4, 0)
                           + '\n')
        compile_file(path,
                     quiet=2)
    except Exception as error:
        warnings.warn(f'Failed saving "{path}". '
                      f'Reason:\n{pretty.format_exception(error)}',
                      UserWarning)



import builtins
import typing as t

from . import catalog
from .arboreal.kind import NodeKind

ModuleAstNodesKinds = t.Mapping[catalog.Path, NodeKind]
ModuleReferences = t.Mapping[catalog.Path, catalog.QualifiedPath]
ModuleSubmodules = t.Collection[catalog.Path]
ModuleSuperclasses = t.Mapping[catalog.Path,
                               t.Collection[catalog.QualifiedPath]]
Scope = t.Mapping[str, 'Scope']


class ObjectNotFound(Exception):
    pass


def scope_contains_path(scope: Scope, path: catalog.Path) -> bool:
    for part in path:
        try:
            scope = scope[part]
        except KeyError:
            return False
    return True


def contains_object_path(
        module_path: catalog.Path,
        parent_path: catalog.Path,
        object_path: catalog.Path,
        modules_definitions: t.Mapping[catalog.Path, Scope],
        modules_references: t.Mapping[catalog.Path, ModuleReferences],
        modules_submodules: t.Mapping[catalog.Path, ModuleSubmodules],
        modules_superclasses: t.Mapping[catalog.Path, ModuleSuperclasses]
) -> bool:
    try:
        resolve_object_path(
                module_path, parent_path, object_path, modules_definitions,
                modules_references, modules_submodules, modules_superclasses
        )
    except ObjectNotFound:
        return False
    else:
        return True


def resolve_object_path(
        module_path: catalog.Path,
        parent_path: catalog.Path,
        object_path: catalog.Path,
        modules_definitions: t.Mapping[catalog.Path, Scope],
        modules_references: t.Mapping[catalog.Path, ModuleReferences],
        modules_submodules: t.Mapping[catalog.Path, ModuleSubmodules],
        modules_superclasses: t.Mapping[catalog.Path, ModuleSuperclasses],
        *visited_modules_paths: catalog.Path,
        _builtins_module_path: catalog.Path
        = catalog.module_path_from_module(builtins)
) -> catalog.QualifiedPath:
    try:
        module_definitions = modules_definitions[module_path]
    except KeyError:
        raise ObjectNotFound((module_path, object_path))
    if not object_path:
        return (module_path, object_path)
    assert scope_contains_path(module_definitions, parent_path)
    if (parent_path and scope_contains_path(module_definitions,
                                            (*parent_path, object_path[0]))):
        return resolve_object_path(
                module_path, (), parent_path + object_path,
                modules_definitions, modules_references, modules_submodules,
                modules_superclasses, *visited_modules_paths
        )
    elif object_path[0] in module_definitions:
        module_superclasses, scope = (
            modules_superclasses.get(module_path, {}),
            module_definitions[object_path[0]]
        )
        for index, sub_name in enumerate(object_path[1:],
                                         start=1):
            if sub_name in scope:
                scope = scope[sub_name]
            else:
                sub_path = object_path[:index]
                try:
                    sub_path_superclasses = module_superclasses[sub_path]
                except KeyError:
                    pass
                else:
                    for (
                            superclass_module_path, superclass_object_path
                    ) in sub_path_superclasses:
                        try:
                            return resolve_object_path(
                                    superclass_module_path, (),
                                    superclass_object_path
                                    + object_path[index:],
                                    modules_definitions, modules_references,
                                    modules_submodules, modules_superclasses,
                                    *visited_modules_paths, module_path
                            )
                        except ObjectNotFound:
                            continue
                if object_path[0] != object.__name__:
                    return resolve_object_path(
                            _builtins_module_path, (),
                            (object.__name__,) + object_path[index:],
                            modules_definitions, modules_references,
                            modules_submodules, modules_superclasses,
                            *visited_modules_paths
                    )
                else:
                    raise ObjectNotFound((module_path, object_path))
        return (module_path, object_path)
    else:
        for sub_module_path in modules_submodules.get(module_path, []):
            if sub_module_path not in visited_modules_paths:
                try:
                    return resolve_object_path(
                            sub_module_path, (), object_path,
                            modules_definitions, modules_references,
                            modules_submodules, modules_superclasses,
                            *visited_modules_paths, module_path
                    )
                except ObjectNotFound:
                    continue
        module_references = modules_references[module_path]
        for offset in range(len(object_path)):
            sub_object_path = object_path[:len(object_path) - offset]
            try:
                referent_module_path, referent_object_path = module_references[
                    sub_object_path
                ]
            except KeyError:
                continue
            else:
                try:
                    referent_module_path, referent_object_path = (
                        resolve_object_path(
                                referent_module_path, (), referent_object_path,
                                modules_definitions, modules_references,
                                modules_submodules, modules_superclasses,
                                *visited_modules_paths, module_path
                        )
                    )
                except ObjectNotFound:
                    assert (len(referent_object_path) == 1), (module_path,
                                                              object_path)
                    referent_module_path += referent_object_path
                    referent_object_path = ()
                    assert (
                            referent_module_path in modules_definitions
                    ), (module_path, object_path)
                return resolve_object_path(
                        referent_module_path, (),
                        referent_object_path
                        + object_path[len(object_path) - offset:],
                        modules_definitions, modules_references,
                        modules_submodules, modules_superclasses,
                        *visited_modules_paths, module_path
                )
        if scope_contains_path(modules_definitions[_builtins_module_path],
                               object_path):
            return _builtins_module_path, object_path
        raise ObjectNotFound((module_path, object_path))



import sys
import types
import typing as t
from collections import abc
from functools import singledispatch

from typing_extensions import (Literal,
                               ParamSpec,
                               ParamSpecArgs,
                               ParamSpecKwargs)

_T = t.TypeVar('_T')
GenericAlias: t.Any = type(t.Generic[_T])
del _T


@singledispatch
def are_equal(left: t.Any, right: t.Any) -> bool:
    result = left == right
    assert isinstance(result, bool), result
    return result


@are_equal.register(GenericAlias)
@are_equal.register(type(t.Union[int, None]))
@are_equal.register(type)
def _(left: t.Union[GenericAlias, t.Type[t.Any]], right: t.Any) -> bool:
    left_args, right_args = to_arguments(left), to_arguments(right)
    left_origin, right_origin = (to_origin(left) or left,
                                 to_origin(right) or right)
    return (left_origin is right_origin
            and len(left_args) == len(right_args)
            and ((all(any(are_equal(left_arg, right_arg)
                          for right_arg in right_args)
                      for left_arg in left_args)
                  and all(any(are_equal(left_arg, right_arg)
                              for left_arg in left_args)
                          for right_arg in right_args))
                 if (left_origin is t.Union or left_origin is Literal)
                 else all(map(are_equal, left_args, right_args))))


@are_equal.register(abc.Sequence)
def _(left: t.Sequence[t.Any], right: t.Any) -> bool:
    return (type(left) is type(right)
            and len(left) == len(right)
            and all(map(are_equal, left, right)))


@are_equal.register(ParamSpec)
def _(left: ParamSpec, right: t.Any) -> bool:
    return (type(left) is type(right)
            and left.__name__ == right.__name__
            and are_equal(left.__bound__, right.__bound__)
            and left.__contravariant__ is right.__contravariant__
            and left.__covariant__ is right.__covariant__)


@are_equal.register(ParamSpecArgs)
def _(left: ParamSpecArgs, right: t.Any) -> bool:
    return (type(left) is type(right)
            and are_equal(left.__origin__, right.__origin__))


if sys.version_info < (3, 10):
    @are_equal.register(types.LambdaType)
    def _(left: types.LambdaType,
          right: t.Any,
          *,
          _sentinel: t.Any = object()) -> bool:
        left_supertype = getattr(left, '__supertype__', _sentinel)
        right_supertype = getattr(right, '__supertype__', _sentinel)
        return (type(left) is type(right)
                and are_equal(left_supertype, right_supertype))
else:
    are_equal.register(types.UnionType, are_equal.dispatch(type))


    @are_equal.register(t.NewType)
    def _(left: t.NewType,
          right: t.Any,
          *,
          _sentinel: t.Any = object()) -> bool:
        return (type(left) is type(right)
                and (getattr(left, '__qualname__', _sentinel)
                     == getattr(right, '__qualname__', _sentinel))
                and are_equal(left.__supertype__, right.__supertype__))


@are_equal.register(ParamSpecKwargs)
def _(left: ParamSpecKwargs, right: t.Any) -> bool:
    return (type(left) is type(right)
            and are_equal(left.__origin__, right.__origin__))


@are_equal.register(t.TypeVar)
def _(left: t.TypeVar, right: t.Any) -> bool:
    left_constraints, right_constraints = (
        getattr(left, '__constraints__', ()),
        getattr(right, '__constraints__', ())
    )
    return (type(left) is type(right)
            and left.__name__ == right.__name__
            and are_equal(left.__bound__, right.__bound__)
            and left.__contravariant__ is right.__contravariant__
            and left.__covariant__ is right.__covariant__
            and len(left_constraints) == len(right_constraints)
            and all(map(are_equal, left_constraints,
                        right_constraints)))


@are_equal.register(bytes)
@are_equal.register(str)
def _(left: t.Union[bytes, str], right: t.Any) -> bool:
    result = left == right
    assert isinstance(result, bool), result
    return result


if sys.version_info >= (3, 8):
    to_arguments = t.get_args
    to_origin = t.get_origin
else:
    def to_arguments(annotation: t.Any) -> t.Tuple[t.Any, ...]:
        if isinstance(annotation, GenericAlias):
            result = annotation.__args__
            if (result
                    and to_origin(annotation) is abc.Callable
                    and result[0] is not Ellipsis):
                result = (list(result[:-1]), result[-1])
            assert isinstance(result, tuple)
            return result
        return ()


    def to_origin(annotation: t.Any) -> t.Any:
        if isinstance(annotation, GenericAlias):
            return annotation.__origin__
        elif annotation is t.Generic:
            return t.Generic
        else:
            return None


@singledispatch
def to_repr(value: t.Any) -> str:
    return repr(value)


@to_repr.register(list)
def _(value: t.List[t.Any]) -> str:
    return f'[{", ".join(to_repr(element) for element in value)}]'


@to_repr.register(tuple)
def _(value: t.Tuple[t.Any, ...]) -> str:
    return (f'({to_repr(value[0])},)'
            if len(value) == 1
            else f'({", ".join(to_repr(element) for element in value)})')


@to_repr.register(type)
def _(value: type) -> str:
    if value in (type(None), type(NotImplemented), type(Ellipsis)):
        return f'{type.__qualname__}({value()!r})'
    else:
        args = to_arguments(value)
        result = f'{value.__module__}.{value.__qualname__}'
        return (f'{result}[{", ".join(map(to_repr, args))}]'
                if args
                else result)


@to_repr.register(GenericAlias)
def _(value: GenericAlias) -> str:
    origin = to_origin(value)
    arguments = to_arguments(value)
    return (((f'{to_repr(t.Optional)}'
              f'[{to_repr(arguments[arguments[0] is type(None)])}]')
             if len(arguments) == 2 and type(None) in arguments
             else (f'{to_repr(origin)}'
                   f'[{", ".join(map(to_repr, arguments))}]'))
            if origin is t.Union
            else (((f'{to_repr(origin)}'
                    f'[{", ".join(map(to_repr, arguments))}]')
                   if arguments
                   else f'{to_repr(origin)}[()]')
                  if value._name is None
                  else ((f'{value.__module__}.{value._name}'
                         f'[{", ".join(map(to_repr, arguments))}]')
                        if arguments
                        else f'{value.__module__}.{value._name}[()]')))


@to_repr.register(t.TypeVar)
def _(value: t.TypeVar) -> str:
    arguments = [repr(value.__name__)]
    arguments.extend(map(to_repr, value.__constraints__))
    if value.__bound__ is not None:
        arguments.append(f'bound={to_repr(value.__bound__)}')
    if value.__contravariant__:
        arguments.append('contravariant=True')
    if value.__covariant__:
        arguments.append('covariant=True')
    return f'{to_repr(type(value))}({", ".join(arguments)})'



import sys as _sys
import types as _types
import typing as _t
from functools import singledispatch as _singledispatch

from .utils import decorate_if as _decorate_if

Path = _t.Tuple[str, ...]
QualifiedPath = _t.Tuple[Path, Path]
SEPARATOR = '.'


def path_from_string(value: str) -> Path:
    assert isinstance(value, str), value
    return tuple(value.split('.'))


def path_to_parent(value: Path) -> Path:
    return value[:-1]


def path_to_string(value: Path) -> str:
    return SEPARATOR.join(value)


@_singledispatch
def qualified_path_from(value: _t.Any) -> QualifiedPath:
    return (), ()


@qualified_path_from.register(_types.BuiltinFunctionType)
@_decorate_if(qualified_path_from.register(_types.BuiltinMethodType),
              _sys.implementation.name != 'pypy')
def _(
        value: _t.Union[_types.BuiltinFunctionType, _types.BuiltinMethodType]
) -> QualifiedPath:
    self = value.__self__
    return ((path_from_string(self.__module__),
             path_from_string(value.__qualname__))
            if isinstance(self, type)
            else ((path_from_string(self.__name__
                                    if self.__spec__ is None
                                    else self.__spec__.name),
                   path_from_string(value.__qualname__))
                  if isinstance(self, _types.ModuleType)
                  else ((),
                        path_from_string(value.__qualname__)
                        if self is None
                        else ())))


@qualified_path_from.register(_types.FunctionType)
def _(value: _types.FunctionType) -> QualifiedPath:
    return (()
            if value.__module__ is None
            else path_from_string(value.__module__),
            path_from_string(value.__qualname__))


@_decorate_if(qualified_path_from.register(_types.MemberDescriptorType),
              _sys.implementation.name == 'pypy')
@_decorate_if(qualified_path_from.register(_types.MethodDescriptorType),
              _sys.implementation.name != 'pypy')
@_decorate_if(qualified_path_from.register(_types.MethodWrapperType),
              _sys.implementation.name != 'pypy')
@_decorate_if(qualified_path_from.register(_types.WrapperDescriptorType),
              _sys.implementation.name != 'pypy')
def _(
        value: _t.Union[_types.MemberDescriptorType,
                        _types.MethodDescriptorType, _types.MethodWrapperType,
                        _types.WrapperDescriptorType]
) -> QualifiedPath:
    return (path_from_string(value.__objclass__.__module__),
            path_from_string(value.__qualname__))


@_decorate_if(qualified_path_from.register(_types.MemberDescriptorType),
              _sys.implementation.name == 'pypy')
def _(value: _types.MemberDescriptorType) -> QualifiedPath:
    return (path_from_string(value.__objclass__.__module__),
            (*path_from_string(value.__objclass__.__qualname__),
             value.__name__))


@qualified_path_from.register(type)
def _(value: type) -> QualifiedPath:
    return (path_from_string(value.__module__),
            path_from_string(value.__qualname__))


def is_attribute(path: Path) -> bool:
    return len(path) > 1


def module_path_from_module(object_: _types.ModuleType) -> Path:
    return path_from_string(object_.__name__
                            if object_.__spec__ is None
                            else object_.__spec__.name)


def object_path_from_callable(value: _t.Callable[..., _t.Any]) -> Path:
    _, object_path = qualified_path_from(value)
    return object_path



import sys
import typing as t

from typing_extensions import Final

from . import catalog
from .sources import stdlib_modules_paths


def _recursively_update_modules_paths(
        set_: t.Set[catalog.Path],
        *names: str,
) -> None:
    paths = [catalog.path_from_string(name) for name in names]
    set_.update({*paths,
                 *[candidate
                   for candidate in stdlib_modules_paths
                   if any(candidate[:len(path)] == path for path in paths)]})


# importing will cause unwanted side effects such as raising error
unsupported_stdlib_modules_paths: t.Set[catalog.Path] = set()
_recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                  'antigravity',
                                  'crypt',
                                  'idlelib',
                                  'lib2to3.pgen2.conv',
                                  'this',
                                  'tkinter',
                                  'turtle',
                                  'turtledemo')

if sys.implementation.name == 'pypy':
    _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                      '__decimal',
                                      '_crypt',
                                      '_curses_build',
                                      '_scproxy',
                                      'future_builtins',
                                      'identity_dict',
                                      'msilib',
                                      'symtable',
                                      'tracemalloc')
    unsupported_stdlib_modules_paths.update(
            [path
             for path in stdlib_modules_paths
             if (path[0].startswith(('__pypy', '_pypy'))
                 or '__pycache__' in path)]
    )
    if sys.platform == 'win32':
        _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                          '_curses',
                                          '_curses_panel',
                                          '_dbm',
                                          '_gdbm',
                                          '_posixshmem',
                                          '_resource_build',
                                          '_sqlite3_build',
                                          '_sysconfigdata',
                                          '_tkinter',
                                          'asyncio.unix_events',
                                          'cffi._pycparser._build_tables',
                                          'ctypes_support',
                                          'curses',
                                          'dbm.gnu',
                                          'dbm.ndbm',
                                          'distutils.command.bdist_msi',
                                          'distutils.command.bdist_wininst',
                                          'distutils.sysconfig_cpython',
                                          'distutils.sysconfig_pypy',
                                          'grp',
                                          'multiprocessing.popen_fork',
                                          'multiprocessing.popen_forkserver',
                                          'multiprocessing.popen_spawn_posix',
                                          'pty',
                                          'pyrepl._minimal_curses',
                                          'pyrepl.curses',
                                          'pyrepl.fancy_termios',
                                          'pyrepl.keymaps',
                                          'pyrepl.pygame_console',
                                          'pyrepl.pygame_keymap',
                                          'pyrepl.readline',
                                          'pyrepl.simple_interact',
                                          'pyrepl.unix_console',
                                          'pyrepl.unix_eventqueue',
                                          'readline',
                                          'resource',
                                          'syslog',
                                          'tty')
    else:
        _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                          '_overlapped',
                                          '_tkinter',
                                          '_winapi',
                                          'asyncio.windows_events',
                                          'asyncio.windows_utils',
                                          'cffi._pycparser._build_tables',
                                          'distutils._msvccompiler',
                                          'distutils.command.bdist_wininst',
                                          'distutils.command.bdist_msi',
                                          'distutils.msvc9compiler',
                                          'distutils.sysconfig_cpython',
                                          'distutils.sysconfig_pypy',
                                          'encodings.mbcs',
                                          'encodings.oem',
                                          'msvcrt',
                                          'multiprocessing.popen_spawn_win32',
                                          'pyrepl.keymaps',
                                          'pyrepl.pygame_console',
                                          'pyrepl.pygame_keymap')
    if sys.version_info >= (3, 9):
        _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                          '_cffi_ssl._cffi_src.build_openssl',
                                          '_ssl_build')
else:
    _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                      'distutils.command.bdist_msi')

    if sys.version_info < (3, 10):
        _recursively_update_modules_paths(
                unsupported_stdlib_modules_paths,
                'distutils.command.bdist_wininst'
        )

    if sys.platform == 'win32':
        _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                          'curses',
                                          'pty',
                                          'tty')
        _recursively_update_modules_paths(
                unsupported_stdlib_modules_paths,
                'asyncio.unix_events',
                'dbm.gnu',
                'dbm.ndbm',
                'multiprocessing.popen_fork',
                'multiprocessing.popen_forkserver',
                'multiprocessing.popen_spawn_posix'
        )
    else:
        _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                          'asyncio.windows_events',
                                          'asyncio.windows_utils',
                                          'ctypes.wintypes',
                                          'distutils.msvc9compiler',
                                          'encodings.mbcs',
                                          'encodings.oem',
                                          'multiprocessing.popen_spawn_win32')
        if sys.version_info < (3, 8):
            _recursively_update_modules_paths(unsupported_stdlib_modules_paths,
                                              'distutils._msvccompiler',
                                              'encodings.cp65001')
        if sys.platform == 'darwin':
            if sys.version_info >= (3, 8):
                _recursively_update_modules_paths(
                        unsupported_stdlib_modules_paths,
                        'dbm.gnu',
                        'distutils._msvccompiler',
                )

supported_stdlib_modules_paths: Final[t.Iterable[catalog.Path]] = [
    module_path
    for module_path in stdlib_modules_paths
    if module_path not in unsupported_stdlib_modules_paths
]



import types
import typing as t
from functools import reduce

from . import catalog

ModuleOrType = t.Union[types.ModuleType, type]
Namespace = t.MutableMapping[str, t.Any]


class ObjectNotFound(Exception):
    pass


def search(module_or_type: ModuleOrType, object_path: catalog.Path) -> t.Any:
    try:
        return reduce(getattr, object_path, module_or_type)
    except AttributeError:
        raise ObjectNotFound(object_path)



import ast
import sys
import typing as t
from functools import singledispatch
from itertools import chain

from paradigm._core import catalog

RawAstNode = t.NewType('RawAstNode', str)


@singledispatch
def to_names(ast_node: ast.AST) -> t.List[str]:
    raise TypeError(type(ast_node))


@to_names.register(ast.AsyncFunctionDef)
@to_names.register(ast.FunctionDef)
@to_names.register(ast.ClassDef)
def _(
        ast_node: t.Union[ast.AsyncFunctionDef, ast.FunctionDef, ast.ClassDef]
) -> t.List[str]:
    return [ast_node.name]


@to_names.register(ast.AnnAssign)
def _(ast_node: ast.AnnAssign) -> t.List[str]:
    return to_names(ast_node.target)


@to_names.register(ast.Assign)
def _(ast_node: ast.Assign) -> t.List[str]:
    return list(chain.from_iterable(to_names(target)
                                    for target in ast_node.targets))


@to_names.register(ast.Name)
def _(ast_node: ast.Name) -> t.List[str]:
    assert isinstance(ast_node.ctx, ast.Store), ast_node
    return [ast_node.id]


@singledispatch
def to_identifier(ast_node: ast.AST) -> str:
    raise TypeError(type(ast_node))


@to_identifier.register(ast.Subscript)
def _(ast_node: ast.Subscript) -> str:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return (to_identifier(ast_node.value)
            + '_getitem_'
            + to_identifier(ast_node.slice))


@to_identifier.register(ast.Ellipsis)
def _(ast_node: ast.Subscript) -> str:
    return repr(Ellipsis)


@to_identifier.register(ast.Constant)
@to_identifier.register(ast.NameConstant)
def _(ast_node: ast.NameConstant) -> str:
    return repr(ast_node.value)


@to_identifier.register(ast.BitAnd)
def _(ast_node: ast.BitAnd) -> str:
    return 'bitand'


@to_identifier.register(ast.BitOr)
def _(ast_node: ast.BitOr) -> str:
    return 'bitor'


@to_identifier.register(ast.BitXor)
def _(ast_node: ast.BitXor) -> str:
    return 'bitxor'


@to_identifier.register(ast.BinOp)
def _(ast_node: ast.BinOp) -> str:
    return (to_identifier(ast_node.left)
            + '_' + to_identifier(ast_node.op) + '_'
            + to_identifier(ast_node.right))


@to_identifier.register(ast.Tuple)
def _(ast_node: ast.Tuple) -> str:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return '_'.join(to_identifier(element) for element in ast_node.elts)


@to_identifier.register(ast.Name)
def _(ast_node: ast.Name) -> str:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return ast_node.id


@to_identifier.register(ast.Attribute)
def _(ast_node: ast.Attribute) -> str:
    assert isinstance(ast_node.ctx, ast.Load), ast_node
    return to_identifier(ast_node.value) + '_' + ast_node.attr


if sys.version_info < (3, 9):
    @to_identifier.register(ast.Index)
    def _(ast_node: ast.Index) -> str:
        return to_identifier(ast_node.value)


def to_raw(ast_node: ast.AST) -> RawAstNode:
    return RawAstNode(ast.dump(ast_node,
                               annotate_fields=False))


@singledispatch
def to_str(ast_node: ast.AST) -> str:
    raise TypeError(type(ast_node))


@to_str.register(ast.Name)
def _(ast_node: ast.Name) -> str:
    return ast_node.id


@to_str.register(ast.BitOr)
def _(ast_node: ast.BitOr) -> str:
    return '|'


@to_str.register(ast.BitAnd)
def _(ast_node: ast.BitAnd) -> str:
    return '&'


@to_str.register(ast.BitXor)
def _(ast_node: ast.BitXor) -> str:
    return '^'


@to_str.register(ast.UAdd)
def _(ast_node: ast.UAdd) -> str:
    return '+'


@to_str.register(ast.USub)
def _(ast_node: ast.USub) -> str:
    return '-'


@to_str.register(ast.UnaryOp)
def _(ast_node: ast.UnaryOp) -> str:
    return f'{to_str(ast_node.op)}({to_str(ast_node.operand)})'


@to_str.register(ast.Constant)
def _(ast_node: ast.Constant) -> str:
    return str(ast_node.value)


@to_str.register(ast.Attribute)
def _(ast_node: ast.Attribute) -> str:
    return f'{to_str(ast_node.value)}.{ast_node.attr}'


@to_str.register(ast.Subscript)
def _(ast_node: ast.Subscript) -> str:
    return (f'{to_str(ast_node.value)}'
            f'[{to_str(ast_node.slice)}]')


@to_str.register(ast.Tuple)
def _(ast_node: ast.Tuple) -> str:
    elements_strings = [to_str(element)
                        for element in ast_node.elts]
    return (f'({elements_strings[0]},)'
            if len(elements_strings) == 1
            else f'({", ".join(elements_strings)})')


@to_str.register(ast.List)
def _(ast_node: ast.List) -> str:
    elements_strings = [to_str(element)
                        for element in ast_node.elts]
    return f'[{", ".join(elements_strings)}]'


if sys.version_info < (3, 8):
    @to_str.register(ast.Ellipsis)
    def _(ast_node: ast.Ellipsis) -> str:
        return str(Ellipsis)

    @to_str.register(ast.NameConstant)
    def _(ast_node: ast.NameConstant) -> str:
        return str(ast_node.value)


    @to_str.register(ast.Num)
    def _(ast_node: ast.Num) -> str:
        return str(ast_node.n)


    @to_str.register(ast.Str)
    def _(ast_node: ast.Str) -> str:
        return ast_node.s

if sys.version_info < (3, 10):
    @to_str.register(ast.BinOp)
    def _(ast_node: ast.BinOp) -> str:
        assert isinstance(ast_node.op, ast.BitOr), ast_node
        left_operand = to_str(ast_node.left)
        right_operand = to_str(ast_node.right)
        return f'{t.__name__}.Union[{left_operand}, {right_operand}]'


    if sys.version_info < (3, 9):
        @to_str.register(ast.Index)
        def _(ast_node: ast.Index) -> str:
            return to_str(ast_node.value)
else:
    @to_str.register(ast.BinOp)
    def _(ast_node: ast.BinOp) -> str:
        left_operand = to_str(ast_node.left)
        right_operand = to_str(ast_node.right)
        return f'{left_operand} {to_str(ast_node.op)} {right_operand}'


@singledispatch
def to_path(ast_node: ast.expr) -> catalog.Path:
    raise TypeError(type(ast_node))


@to_path.register(ast.Name)
def _(ast_node: ast.Name) -> catalog.Path:
    return (ast_node.id,)


@to_path.register(ast.Attribute)
def _(ast_node: ast.Attribute) -> catalog.Path:
    return to_path(ast_node.value) + (ast_node.attr,)


@singledispatch
def to_maybe_path(ast_node: ast.expr) -> t.Optional[catalog.Path]:
    return None


@to_maybe_path.register(ast.Name)
def _(ast_node: ast.Name) -> t.Optional[catalog.Path]:
    return (ast_node.id,)


@to_maybe_path.register(ast.Attribute)
def _(ast_node: ast.Attribute) -> t.Optional[catalog.Path]:
    value_maybe_path = to_maybe_path(ast_node.value)
    return (None
            if value_maybe_path is None
            else (*value_maybe_path, ast_node.attr))



import typing as _t

from ._core import (models as _models,
                    signatures as _signatures)

OptionalParameter = _models.OptionalParameter
OverloadedSignature = _models.OverloadedSignature
ParameterKind = _models.ParameterKind
PlainSignature = _models.PlainSignature
RequiredParameter = _models.RequiredParameter


def signature_from_callable(
        _callable: _t.Callable[..., _t.Any]
) -> _t.Union[OverloadedSignature, PlainSignature]:
    return _signatures.from_callable(_callable)



import inspect
import sys
import sysconfig
import typing as t
from importlib.machinery import (EXTENSION_SUFFIXES,
                                 SOURCE_SUFFIXES)
from importlib.util import find_spec
from itertools import chain
from pathlib import Path as _Path

from . import (catalog,
               file_system)

_STUB_EXTENSION = '.pyi'

Path = _Path


class NotFound(Exception):
    pass


def from_module_path(module_path: catalog.Path) -> Path:
    try:
        return _stubs_cache[module_path]
    except KeyError as error:
        raise NotFound(module_path) from error


def is_package(module_path: catalog.Path) -> bool:
    try:
        source_path = from_module_path(module_path)
    except NotFound:
        spec = find_spec(catalog.path_to_string(module_path))
        if spec is None or spec.origin is None:
            return False
        source_path = Path(spec.origin)
    return source_path.stem == file_system.INIT_MODULE_NAME


def _find_source_path(module_name: str) -> Path:
    maybe_spec = find_spec(module_name)
    assert maybe_spec is not None
    maybe_path_string = maybe_spec.origin
    assert maybe_path_string is not None
    return Path(maybe_path_string)


def _to_stubs_cache(
        root: Path = _find_source_path('mypy').parent / 'typeshed' / 'stdlib'
) -> t.Dict[catalog.Path, Path]:
    assert root.exists(), root

    def to_module_path(stub_path: Path) -> catalog.Path:
        return _relative_file_path_to_module_path(
                stub_path.relative_to(root).with_suffix('.py')
        )

    return {to_module_path(file_path): file_path
            for file_path in file_system.find_files_paths(root)
            if _is_stub(file_path)}


def _is_stub(path: Path) -> bool:
    return path.suffixes == [_STUB_EXTENSION]


def _relative_file_path_to_module_path(path: Path) -> catalog.Path:
    assert not path.is_absolute(), 'Path should be relative.'
    *parent_path_parts, module_file_name = path.parts
    parent_path = tuple(parent_path_parts)
    module_name = inspect.getmodulename(module_file_name)
    return (parent_path
            if (module_name is None
                or module_name == file_system.INIT_MODULE_NAME)
            else parent_path + (module_name,))


_stubs_cache = _to_stubs_cache()
stubs_stdlib_modules_paths = set(_stubs_cache.keys())
_sources_directories = {
    Path(sysconfig.get_path('platstdlib')), Path(sysconfig.get_path('stdlib'))
}


def _to_modules_paths(root: Path) -> t.Iterable[catalog.Path]:
    assert root.exists(), root

    def is_source_path(
            path: Path,
            *,
            _suffixes: t.Container[str] = frozenset(SOURCE_SUFFIXES
                                                    + EXTENSION_SUFFIXES)
    ) -> bool:
        return ''.join(path.suffixes) in _suffixes

    def to_module_path(source_path: Path) -> catalog.Path:
        return _relative_file_path_to_module_path(
                source_path.relative_to(root)
        )

    return {to_module_path(file_path)
            for file_path in file_system.find_files_paths(root)
            if is_source_path(file_path)}


def _is_valid_module_path(module_path: catalog.Path) -> bool:
    return (bool(module_path)
            and 'test' not in module_path
            and 'tests' not in module_path
            and module_path[-1] != '__main__'
            and all((part.isidentifier()
                     and not part.startswith(('_test', 'test_'))
                     and not part.endswith('_test'))
                    for part in module_path))


stdlib_modules_paths = dict.fromkeys(
        chain([catalog.path_from_string(module_name)
               for module_name in sys.builtin_module_names],
              [module_path
               for path in _sources_directories
               for module_path in _to_modules_paths(path)
               if _is_valid_module_path(module_path)])
).keys()



import sys as _sys
import typing as _t
from importlib import import_module as _import_module
from pathlib import Path as _Path

import mypy as _mypy
from mypy.version import __version__ as _mypy_version

from paradigm import __version__ as _version
from . import (caching as _caching,
               catalog as _catalog,
               scoping as _scoping)
from .arboreal import conversion as _serialization
from .arboreal.kind import NodeKind as _NodeKind

_CACHE_PATH = _Path(__file__).with_name(
        '_' + _mypy.__name__ + '_' + _mypy_version.replace('.', '_')
        + '_' + _sys.platform
        + '_' + _sys.implementation.name
        + '_' + '_'.join(map(str, _sys.version_info))
        + '_' + _Path(__file__).with_suffix(_caching.FILE_SUFFIX).name
)
_DEFINITIONS_FIELD_NAME = 'definitions'
_NODES_KINDS_FIELD_NAME = 'nodes_kinds'
_RAW_AST_NODES_FIELD_NAME = 'raw_ast_nodes'
_REFERENCES_FIELD_NAME = 'references'
_SUBMODULES_FIELD_NAME = 'submodules'
_SUPERCLASSES_FIELD_NAME = 'superclasses'
_VERSION_FIELD_NAME = 'version'

ObjectRawAstNodes = _t.List[_serialization.RawAstNode]
_ModuleRawAstNodes = _t.Dict[_catalog.Path, ObjectRawAstNodes]

definitions: _t.Dict[_catalog.Path, _scoping.Scope]
nodes_kinds: _t.Dict[_catalog.Path, _scoping.ModuleAstNodesKinds]
raw_ast_nodes: _t.Dict[_catalog.Path, _ModuleRawAstNodes]
references: _t.Dict[_catalog.Path, _scoping.ModuleReferences]
submodules: _t.Dict[_catalog.Path, _scoping.ModuleSubmodules]
superclasses: _t.Dict[_catalog.Path, _scoping.ModuleSuperclasses]

try:
    (
        definitions, _raw_nodes_kinds, raw_ast_nodes, references, submodules,
        superclasses, _cached_version
    ) = _caching.load(_DEFINITIONS_FIELD_NAME, _NODES_KINDS_FIELD_NAME,
                      _RAW_AST_NODES_FIELD_NAME, _REFERENCES_FIELD_NAME,
                      _SUBMODULES_FIELD_NAME, _SUPERCLASSES_FIELD_NAME,
                      _VERSION_FIELD_NAME,
                      path=_CACHE_PATH)
except Exception:
    _reload_cache = True
else:
    _reload_cache = _cached_version != _version
    if not _reload_cache:
        nodes_kinds = {
            module_path: {
                object_path: _NodeKind(raw_node_kind)
                for object_path, raw_node_kind in objects_kinds.items()
            }
            for module_path, objects_kinds in _raw_nodes_kinds.items()
        }
if _reload_cache:
    import ast as _ast
    import builtins as _builtins
    from copy import deepcopy as _deepcopy
    from functools import singledispatch as _singledispatch
    from itertools import chain

    from typing_extensions import TypeGuard as _TypeGuard

    from . import (execution as _execution,
                   namespacing as _namespacing,
                   sources as _sources)
    from .arboreal import (construction as _construction,
                           conversion as _conversion)
    from .arboreal.execution import execute_statement as _execute_statement
    from .arboreal.utils import (
        is_dependency_name as _is_dependency_name,
        recursively_iterate_children as _recursively_iterate_children,
        subscript_to_item as _subscript_to_item,
        to_parent_module_path as _to_parent_module_path
    )
    from .sources import stubs_stdlib_modules_paths as _stdlib_modules_paths

    _ObjectAstNodes = _t.List[_ast.AST]
    _ModuleAstNodes = _t.Dict[_catalog.Path, _ObjectAstNodes]
    _ModuleAstNodesKinds = _t.Dict[_catalog.Path, _NodeKind]
    _ModuleReferences = _t.Dict[_catalog.Path, _catalog.QualifiedPath]
    _ModuleSubmodules = _t.List[_catalog.Path]
    _ModuleSuperclasses = _t.Dict[
        _catalog.Path, _t.List[_catalog.QualifiedPath]
    ]
    _Scope = _t.Dict[str, '_Scope']


    def _named_tuple_to_constructor_ast_node(
            ast_node: _ast.ClassDef
    ) -> _ast.FunctionDef:
        annotations_ast_nodes: _t.List[_ast.AnnAssign] = [
            ast_child_node
            for ast_child_node in ast_node.body
            if isinstance(ast_child_node, _ast.AnnAssign)
        ]
        assert all(isinstance(ast_node.target, _ast.Name)
                   for ast_node in annotations_ast_nodes), ast_node
        return _ast.FunctionDef(
                '__new__',
                _annotations_to_signature(annotations_ast_nodes),
                [_ast.Expr(_ast.Ellipsis())], [],
                _ast.Name(ast_node.name, _ast.Load())
        )


    if _sys.version_info < (3, 8):
        def _annotations_to_signature(
                ast_nodes: _t.List[_ast.AnnAssign]
        ) -> _ast.arguments:
            return _ast.arguments([_ast.arg('cls', None)]
                                  + [_ann_assign_to_arg(ast_node)
                                     for ast_node in ast_nodes],
                                  None, [], [], None,
                                  [ast_node.value
                                   for ast_node in ast_nodes
                                   if ast_node.value is not None])
    else:
        def _annotations_to_signature(
                ast_nodes: _t.List[_ast.AnnAssign]
        ) -> _ast.arguments:
            return _ast.arguments([],
                                  [_ast.arg('cls', None)]
                                  + [_ann_assign_to_arg(ast_node)
                                     for ast_node in ast_nodes],
                                  None, [], [], None,
                                  [ast_node.value
                                   for ast_node in ast_nodes
                                   if ast_node.value is not None])


    def _ann_assign_to_arg(ast_node: _ast.AnnAssign) -> _ast.arg:
        assert isinstance(ast_node.target, _ast.Name), ast_node
        return _ast.arg(ast_node.target.id, ast_node.annotation)


    def _evaluate_expression(node: _ast.expr,
                             *,
                             source_path: _Path,
                             namespace: _namespacing.Namespace) -> _t.Any:
        # to avoid name conflicts
        # we're using name that won't be present
        # because it'll lead to ``SyntaxError`` otherwise
        # and no AST will be generated
        temporary_name = '@tmp'
        assignment = _expression_to_assignment(node,
                                               name=temporary_name)
        _execute_statement(assignment,
                           source_path=source_path,
                           namespace=namespace)
        return namespace.pop(temporary_name)


    def _expression_to_assignment(node: _ast.expr,
                                  *,
                                  name: str) -> _ast.Assign:
        name_node = _ast.copy_location(_ast.Name(name, _ast.Store()), node)
        return _ast.copy_location(_ast.Assign([name_node], node), node)


    def _is_generic_specialization(
            base: _ast.expr
    ) -> _TypeGuard[_ast.Subscript]:
        return isinstance(base, _ast.Subscript)


    class _StateParser(_ast.NodeVisitor):
        def __init__(
                self,
                module_path: _catalog.Path,
                parent_path: _catalog.Path,
                source_path: _Path,
                scope_definitions: _Scope,
                module_ast_nodes: _ModuleAstNodes,
                module_ast_nodes_kinds: _ModuleAstNodesKinds,
                module_references: _ModuleReferences,
                modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
                modules_ast_nodes_kinds: _t.Dict[
                    _catalog.Path, _ModuleAstNodesKinds
                ],
                modules_definitions: _t.Dict[_catalog.Path, _Scope],
                modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
                modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules],
                modules_superclasses: _t.Dict[
                    _catalog.Path, _ModuleSuperclasses
                ],
                visited_modules_paths: _t.Set[_catalog.Path],
                classes_bases: _t.Dict[
                    _catalog.QualifiedPath, _t.List[_ast.expr]
                ],
                generics_parameters_paths: _t.Dict[
                    _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
                ]
        ) -> None:
            (
                self.generics_parameters_paths, self.module_ast_nodes,
                self.module_ast_nodes_kinds, self.module_path,
                self.module_references, self.modules_ast_nodes,
                self.modules_ast_nodes_kinds, self.modules_definitions,
                self.modules_references, self.modules_submodules,
                self.modules_superclasses, self.parent_path,
                self.scope_definitions, self.source_path, self.classes_bases,
                self.visited_modules_paths
            ) = (
                generics_parameters_paths, module_ast_nodes,
                module_ast_nodes_kinds, module_path, module_references,
                modules_ast_nodes, modules_ast_nodes_kinds,
                modules_definitions, modules_references, modules_submodules,
                modules_superclasses, parent_path, scope_definitions,
                source_path, classes_bases, visited_modules_paths
            )

        def visit_AnnAssign(self, node: _ast.AnnAssign) -> None:
            target_path = _conversion.to_path(node.target)
            value_ast_node = node.value
            value_path = (None
                          if value_ast_node is None
                          else _conversion.to_maybe_path(value_ast_node))
            if value_path is None:
                self._add_ast_node(target_path, node)
                self._add_ast_node_kind(target_path,
                                        _NodeKind.ANNOTATED_ASSIGNMENT)
                self._add_path_definition(target_path)
            else:
                value_module_path, value_object_path = self._to_qualified_path(
                        value_path
                )
                self._add_reference(target_path, value_module_path,
                                    value_object_path)

        def visit_Assign(self, node: _ast.Assign) -> None:
            value_path = _conversion.to_maybe_path(node.value)
            if value_path is None:
                for target in node.targets:
                    target_path = _conversion.to_path(target)
                    self._add_ast_node(target_path, node)
                    self._add_ast_node_kind(target_path, _NodeKind.ASSIGNMENT)
                    self._add_path_definition(target_path)
            else:
                value_module_path, value_object_path = self._to_qualified_path(
                        value_path
                )
                for target in node.targets:
                    target_path = _conversion.to_path(target)
                    self._add_reference(target_path, value_module_path,
                                        value_object_path)

        def visit_AsyncFunctionDef(self, node: _ast.AsyncFunctionDef) -> None:
            function_name = node.name
            self._add_ast_node((function_name,), node)
            self._add_ast_node_kind((function_name,), _NodeKind.ASYNC_FUNCTION)
            self._add_name_definition(function_name)

        def visit_ClassDef(
                self,
                node: _ast.ClassDef,
                *,
                builtins_module_path: _catalog.Path
                = _catalog.module_path_from_module(_builtins),
                generic_object_path: _catalog.Path = ('Generic',),
                protocol_object_path: _catalog.Path = ('Protocol',),
                named_tuple_object_path: _catalog.Path
                = _catalog.path_from_string(_t.NamedTuple.__qualname__),
                typing_module_path: _catalog.Path
                = _catalog.module_path_from_module(_t)
        ) -> None:
            class_name = node.name
            self._add_ast_node_kind((class_name,), _NodeKind.CLASS)
            self._add_name_definition(class_name)
            class_path, module_path = (self.parent_path + (class_name,),
                                       self.module_path)
            class_bases = self.classes_bases[(module_path, class_path)] = []
            type_vars_local_paths = []
            for base in node.bases:
                if _is_generic_specialization(base):
                    type_args_maybe_objects_paths = [
                        _conversion.to_maybe_path(argument)
                        for argument in _collect_type_args(base)
                    ]
                    type_vars_local_paths += [
                        maybe_object_path
                        for maybe_object_path in type_args_maybe_objects_paths
                        if (maybe_object_path is not None
                            and
                            self._is_type_var_object_path(maybe_object_path))
                    ]
                    base_origin_maybe_object_path = (
                        _conversion.to_maybe_path(base.value)
                    )
                    if (base_origin_maybe_object_path is not None
                            and self._resolve_object_path(
                                    base_origin_maybe_object_path
                            ) in [(typing_module_path, generic_object_path),
                                  (typing_module_path, protocol_object_path)]):
                        continue
                else:
                    base_reference_path = _conversion.to_path(base)
                    assert (
                            base_reference_path is not None
                    ), (module_path, class_path)
                    base_module_path, base_object_path = (
                        self._to_qualified_path(base_reference_path)
                    )
                    if (base_module_path == typing_module_path
                            and base_object_path == named_tuple_object_path):
                        constructor_node = (
                            _named_tuple_to_constructor_ast_node(node)
                        )
                        self._add_ast_node((class_name,), constructor_node)
                        self._add_superclass(class_path, base_module_path,
                                             base_object_path)
                        continue
                    elif (base_module_path == typing_module_path
                          and base_object_path == protocol_object_path):
                        continue
                    assert (
                        not (base_module_path == typing_module_path
                             and base_object_path == generic_object_path)
                    ), (module_path, class_path)
                class_bases.append(base)
            generic_parameters = tuple(dict.fromkeys(type_vars_local_paths))
            if generic_parameters:
                self.generics_parameters_paths[
                    (module_path, class_path)
                ] = generic_parameters
            parse_child = _StateParser(
                    module_path, class_path, self.source_path,
                    self.scope_definitions[class_name], self.module_ast_nodes,
                    self.module_ast_nodes_kinds, self.module_references,
                    self.modules_ast_nodes, self.modules_ast_nodes_kinds,
                    self.modules_definitions, self.modules_references,
                    self.modules_submodules, self.modules_superclasses,
                    self.visited_modules_paths, self.classes_bases,
                    self.generics_parameters_paths
            ).visit
            for child in node.body:
                parse_child(child)

        def visit_FunctionDef(self, node: _ast.FunctionDef) -> None:
            function_name = node.name
            self._add_ast_node((function_name,), node)
            self._add_ast_node_kind((function_name,), _NodeKind.FUNCTION)
            self._add_name_definition(function_name)

        def visit_If(self, node: _ast.If) -> None:
            namespace: _namespacing.Namespace = {}
            for dependency_name in {
                child.id
                for child in _recursively_iterate_children(node.test)
                if _is_dependency_name(child)
            }:
                module_path, object_path = self._resolve_object_path(
                        (dependency_name,)
                )
                module = _import_module(_catalog.path_to_string(module_path))
                namespace[dependency_name] = (_namespacing.search(module,
                                                                  object_path)
                                              if object_path
                                              else module)
            condition = _evaluate_expression(node.test,
                                             source_path=self.source_path,
                                             namespace=namespace)
            for child in (node.body if condition else node.orelse):
                self.visit(child)

        def visit_Import(self, node: _ast.Import) -> None:
            for alias in node.names:
                module_path = _catalog.path_from_string(alias.name)
                if alias.asname is None:
                    sub_module_path: _catalog.Path = ()
                    for module_name_part in module_path:
                        sub_module_path += (module_name_part,)
                        self._add_reference(sub_module_path, sub_module_path,
                                            ())
                else:
                    self._add_reference((alias.asname,), module_path, ())

        def visit_ImportFrom(self, node: _ast.ImportFrom) -> None:
            parent_module_path = _to_parent_module_path(
                    node,
                    parent_module_path=self.module_path
            )
            for alias in node.names:
                if alias.name == '*':
                    self._add_submodule(parent_module_path)
                else:
                    actual_path = _catalog.path_from_string(alias.name)
                    self._add_reference(actual_path
                                        if alias.asname is None
                                        else (alias.asname,),
                                        parent_module_path, actual_path)

        def _add_ast_node(self,
                          path: _catalog.Path,
                          ast_node: _ast.AST) -> None:
            self.module_ast_nodes.setdefault(
                    self.parent_path + path, []
            ).append(ast_node)

        def _add_ast_node_kind(self,
                               path: _catalog.Path,
                               node_kind: _NodeKind) -> None:
            self.module_ast_nodes_kinds[self.parent_path + path] = node_kind

        def _add_name_definition(self, name: str) -> None:
            assert isinstance(name, str), name
            self.scope_definitions.setdefault(name, {})

        def _add_path_definition(self, path: _catalog.Path) -> None:
            scope_definitions = self.scope_definitions
            for part in path:
                scope_definitions = scope_definitions.setdefault(part, {})

        def _add_reference(self,
                           reference_path: _catalog.Path,
                           referent_module_path: _catalog.Path,
                           referent_object_path: _catalog.Path) -> None:
            self.module_references[reference_path] = (referent_module_path,
                                                      referent_object_path)

        def _add_submodule(self, module_path: _catalog.Path) -> None:
            self.modules_submodules.setdefault(
                    self.module_path, []
            ).append(module_path)

        def _add_superclass(self,
                            child_path: _catalog.Path,
                            base_module_path: _catalog.Path,
                            base_object_path: _catalog.Path) -> None:
            self.modules_superclasses.setdefault(
                    self.module_path, {}
            ).setdefault(child_path, []).append((base_module_path,
                                                 base_object_path))

        def _resolve_object_path(
                self, object_path: _catalog.Path
        ) -> _catalog.QualifiedPath:
            return _resolve_object_path(
                    self.module_path, self.parent_path, object_path,
                    self.modules_ast_nodes, self.modules_ast_nodes_kinds,
                    self.modules_definitions, self.modules_references,
                    self.modules_submodules, self.modules_superclasses,
                    self.visited_modules_paths, self.classes_bases,
                    self.generics_parameters_paths
            )

        def _to_qualified_path(
                self, object_path: _catalog.Path
        ) -> _catalog.QualifiedPath:
            try:
                return self._resolve_object_path(object_path)
            except _ObjectNotFound:
                self._add_path_definition(object_path)
                return (self.module_path, object_path)

        def _is_type_var_object_path(
                self, object_path: _catalog.Path,
                *,
                type_var_object_path: _catalog.Path
                = _catalog.path_from_string(_t.TypeVar.__qualname__),
                typing_module_path: _catalog.Path
                = _catalog.module_path_from_module(_t),
        ) -> bool:
            module_path, object_path = self._to_qualified_path(object_path)
            _parse_stub_scope(
                    module_path, self.modules_ast_nodes,
                    self.modules_ast_nodes_kinds, self.modules_definitions,
                    self.modules_references, self.modules_submodules,
                    self.modules_superclasses, self.visited_modules_paths,
                    self.classes_bases, self.generics_parameters_paths
            )
            module_ast_nodes = self.modules_ast_nodes[module_path]
            ast_nodes = module_ast_nodes.get(object_path, [])
            if len(ast_nodes) != 1:
                return False
            ast_node, = ast_nodes
            if not (isinstance(ast_node, (_ast.AnnAssign, _ast.Assign))
                    and isinstance(ast_node.value, _ast.Call)):
                return False
            callable_maybe_object_path = _conversion.to_maybe_path(
                    ast_node.value.func
            )
            if callable_maybe_object_path is None:
                return False
            callable_module_path, callable_object_path = (
                _resolve_object_path(
                        module_path, (), callable_maybe_object_path,
                        self.modules_ast_nodes, self.modules_ast_nodes_kinds,
                        self.modules_definitions, self.modules_references,
                        self.modules_submodules, self.modules_superclasses,
                        self.visited_modules_paths, self.classes_bases,
                        self.generics_parameters_paths
                )
            )
            return (callable_module_path == typing_module_path
                    and callable_object_path == type_var_object_path)


    class _ObjectNotFound(Exception):
        pass


    def _resolve_object_path(
            module_path: _catalog.Path,
            parent_path: _catalog.Path,
            object_path: _catalog.Path,
            modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
            modules_ast_nodes_kinds: _t.Dict[
                _catalog.Path, _ModuleAstNodesKinds
            ],
            modules_definitions: _t.Dict[_catalog.Path, _Scope],
            modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
            modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules],
            modules_superclasses: _t.Dict[_catalog.Path, _ModuleSuperclasses],
            visited_modules_paths: _t.Set[_catalog.Path],
            classes_bases: _t.Dict[_catalog.QualifiedPath, _t.List[_ast.expr]],
            generics_parameters_paths: _t.Dict[
                _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
            ],
            *,
            _builtins_module_path: _catalog.Path
            = _catalog.module_path_from_module(_builtins)
    ) -> _catalog.QualifiedPath:
        first_name = object_path[0]
        module_definitions = _parse_stub_scope(
                module_path, modules_ast_nodes, modules_ast_nodes_kinds,
                modules_definitions, modules_references, modules_submodules,
                modules_superclasses, visited_modules_paths, classes_bases,
                generics_parameters_paths
        )
        if _scoping.scope_contains_path(module_definitions,
                                        parent_path + (first_name,)):
            return module_path, parent_path + object_path
        elif first_name in module_definitions:
            return module_path, object_path
        else:
            for sub_module_path in modules_submodules.get(module_path, []):
                try:
                    return _resolve_object_path(
                            sub_module_path, (), object_path,
                            modules_ast_nodes, modules_ast_nodes_kinds,
                            modules_definitions, modules_references,
                            modules_submodules, modules_superclasses,
                            visited_modules_paths, classes_bases,
                            generics_parameters_paths
                    )
                except _ObjectNotFound:
                    continue
        module_references = modules_references[module_path]
        for offset in range(len(object_path)):
            sub_object_path = object_path[:len(object_path) - offset]
            try:
                referent_module_name, referent_object_path = (
                    module_references[sub_object_path]
                )
            except KeyError:
                continue
            else:
                return (referent_module_name,
                        (referent_object_path
                         + object_path[len(object_path) - offset:]))
        if _scoping.scope_contains_path(
                modules_definitions[_builtins_module_path], object_path
        ):
            return _builtins_module_path, object_path
        raise _ObjectNotFound(object_path)


    def _parse_stub_scope(
            module_path: _catalog.Path,
            modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
            modules_ast_nodes_kinds: _t.Dict[
                _catalog.Path, _ModuleAstNodesKinds
            ],
            modules_definitions: _t.Dict[_catalog.Path, _Scope],
            modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
            modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules],
            modules_superclasses: _t.Dict[_catalog.Path, _ModuleSuperclasses],
            visited_modules_paths: _t.Set[_catalog.Path],
            classes_bases: _t.Dict[_catalog.QualifiedPath, _t.List[_ast.expr]],
            generics_parameters_paths: _t.Dict[
                _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
            ]
    ) -> _Scope:
        if module_path in visited_modules_paths:
            return modules_definitions[module_path]
        visited_modules_paths.add(module_path)
        source_path = _sources.from_module_path(module_path)
        module_definitions = modules_definitions[module_path] = {}
        module_ast_nodes_kinds = modules_ast_nodes_kinds[module_path] = {}
        module_ast_nodes = modules_ast_nodes[module_path] = {}
        module_references = modules_references[module_path] = {}
        ast_node = _construction.from_source_path(source_path)
        _StateParser(
                module_path, (), source_path, module_definitions,
                module_ast_nodes, module_ast_nodes_kinds, module_references,
                modules_ast_nodes, modules_ast_nodes_kinds,
                modules_definitions, modules_references, modules_submodules,
                modules_superclasses, visited_modules_paths, classes_bases,
                generics_parameters_paths
        ).visit(ast_node)
        return module_definitions


    @_singledispatch
    def _unpack_ast_node(ast_node: _ast.expr) -> _t.Tuple[_ast.expr, ...]:
        return (ast_node,)


    @_unpack_ast_node.register(_ast.Tuple)
    def _(ast_node: _ast.Tuple) -> _t.Tuple[_ast.expr, ...]:
        assert isinstance(ast_node.ctx, _ast.Load), ast_node
        return tuple(ast_node.elts)


    def _collect_type_args(ast_node: _ast.expr) -> _t.List[_ast.expr]:
        if _is_generic_specialization(ast_node):
            queue = [_subscript_to_item(ast_node)]
            args: _t.List[_ast.expr] = []
            while queue:
                specialization_node = queue.pop()
                candidates = _unpack_ast_node(specialization_node)
                for candidate in reversed(candidates):
                    if _is_generic_specialization(candidate):
                        queue.append(_subscript_to_item(candidate))
                    else:
                        args.append(candidate)
            return args[::-1]
        else:
            return []


    class _SpecializeGeneric(_ast.NodeTransformer):
        def __init__(self, table: _t.Dict[_catalog.Path, _ast.expr]) -> None:
            self.table = table

        def visit_Name(self, node: _ast.Name) -> _ast.expr:
            if not isinstance(node.ctx, _ast.Load):
                return node
            candidate = self.table.get(_conversion.to_path(node))
            return (node
                    if candidate is None
                    else _ast.copy_location(_deepcopy(candidate), node))

        def visit_Attribute(self, node: _ast.Attribute) -> _ast.expr:
            if not isinstance(node.ctx, _ast.Load):
                return node
            candidate = self.table.get(_conversion.to_path(node))
            return (node
                    if candidate is None
                    else _ast.copy_location(_deepcopy(candidate), node))


    def _parse_stubs_state(
            modules_paths: _t.Iterable[_catalog.Path]
    ) -> _t.Tuple[
        _t.Dict[_catalog.Path, _Scope],
        _t.Dict[_catalog.Path, _ModuleAstNodesKinds],
        _t.Dict[_catalog.Path, _ModuleRawAstNodes],
        _t.Dict[_catalog.Path, _ModuleReferences],
        _t.Dict[_catalog.Path, _ModuleSubmodules],
        _t.Dict[_catalog.Path, _ModuleSuperclasses]
    ]:
        modules_definitions: _t.Dict[_catalog.Path, _Scope] = {}
        modules_nodes_kinds: _t.Dict[_catalog.Path, _ModuleAstNodesKinds] = {}
        modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes] = {}
        modules_references: _t.Dict[_catalog.Path, _ModuleReferences] = {}
        modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules] = {}
        modules_superclasses: _t.Dict[_catalog.Path, _ModuleSuperclasses] = {}
        classes_bases: _t.Dict[_catalog.QualifiedPath, _t.List[_ast.expr]] = {}
        generics_parameters_paths: _t.Dict[
            _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
        ] = {}
        visited_modules_paths: _t.Set[_catalog.Path] = set()
        _parse_stub_scope(
                _catalog.module_path_from_module(_builtins), modules_ast_nodes,
                modules_nodes_kinds, modules_definitions, modules_references,
                modules_submodules, modules_superclasses,
                visited_modules_paths, classes_bases, generics_parameters_paths
        )
        for module_path in modules_paths:
            _parse_stub_scope(
                    module_path, modules_ast_nodes, modules_nodes_kinds,
                    modules_definitions, modules_references,
                    modules_submodules, modules_superclasses,
                    visited_modules_paths, classes_bases,
                    generics_parameters_paths
            )
        _process_classes_bases(
                classes_bases, generics_parameters_paths, modules_ast_nodes,
                modules_definitions, modules_nodes_kinds, modules_references,
                modules_submodules, modules_superclasses
        )
        modules_raw_ast_nodes = {
            module_path: {
                object_path: [_serialization.to_raw(ast_node)
                              for ast_node in ast_nodes]
                for object_path, ast_nodes in module_ast_nodes.items()
            }
            for module_path, module_ast_nodes in modules_ast_nodes.items()
        }
        return (
            modules_definitions, modules_nodes_kinds, modules_raw_ast_nodes,
            modules_references, modules_submodules, modules_superclasses
        )


    def _process_classes_bases(
            classes_bases: _t.Dict[_catalog.QualifiedPath, _t.List[_ast.expr]],
            generics_parameters_paths: _t.Dict[
                _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
            ],
            modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
            modules_definitions: _t.Dict[_catalog.Path, _Scope],
            modules_nodes_kinds: _t.Dict[_catalog.Path, _ModuleAstNodesKinds],
            modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
            modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules],
            modules_superclasses: _t.Dict[_catalog.Path, _ModuleSuperclasses],
            *,
            specializations_module_path: _catalog.Path = ('__specializations',)
    ) -> None:
        assert specializations_module_path not in modules_ast_nodes
        specializations_ast_nodes = modules_ast_nodes[
            specializations_module_path
        ] = {}
        assert specializations_module_path not in modules_definitions
        specializations_module_scope = modules_definitions[
            specializations_module_path
        ] = {}
        assert specializations_module_path not in modules_nodes_kinds
        specializations_nodes_kinds = modules_nodes_kinds[
            specializations_module_path
        ] = {}
        assert specializations_module_path not in modules_references
        specializations_references = modules_references[
            specializations_module_path
        ] = {}
        for (class_module_path, class_object_path), class_bases in (
                classes_bases.items()
        ):
            for base in class_bases:
                base_module_path, base_object_path = _register_base_ast_node(
                        base, class_module_path, class_object_path,
                        classes_bases, generics_parameters_paths,
                        specializations_ast_nodes, specializations_module_path,
                        specializations_module_scope,
                        specializations_nodes_kinds,
                        specializations_references, modules_ast_nodes,
                        modules_definitions, modules_nodes_kinds,
                        modules_references, modules_submodules,
                        modules_superclasses
                )
                modules_superclasses.setdefault(
                        class_module_path, {}
                ).setdefault(class_object_path, []).append((base_module_path,
                                                            base_object_path))


    def _register_base_ast_node(
            base: _ast.expr,
            child_module_path: _catalog.Path,
            child_object_path: _catalog.Path,
            classes_bases: _t.Dict[_catalog.QualifiedPath, _t.List[_ast.expr]],
            generics_parameters_paths: _t.Dict[
                _catalog.QualifiedPath, _t.Tuple[_catalog.Path, ...]
            ],
            specializations_ast_nodes: _ModuleAstNodes,
            specializations_module_path: _catalog.Path,
            specializations_module_scope: _Scope,
            specializations_nodes_kinds: _ModuleAstNodesKinds,
            specializations_references: _ModuleReferences,
            modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
            modules_definitions: _t.Dict[_catalog.Path, _Scope],
            modules_nodes_kinds: _t.Dict[_catalog.Path, _ModuleAstNodesKinds],
            modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
            modules_submodules: _t.Dict[_catalog.Path, _ModuleSubmodules],
            modules_superclasses: _t.Dict[_catalog.Path, _ModuleSuperclasses]
    ) -> _catalog.QualifiedPath:
        if _is_generic_specialization(base):
            base_name = _serialization.to_identifier(base)
            base_object_path = (base_name,)
            if base_name not in specializations_module_scope:
                specialization_args = _unpack_ast_node(
                        _subscript_to_item(base)
                )
                generic_object_path = _conversion.to_path(base.value)
                generic_module_path, generic_object_path = (
                    _scoping.resolve_object_path(
                            child_module_path, child_object_path[:-1],
                            generic_object_path, modules_definitions,
                            modules_references, modules_submodules,
                            modules_superclasses
                    )
                )
                generic_parameters_paths = generics_parameters_paths[
                    (generic_module_path, generic_object_path)
                ]
                _register_generic_specialization(
                        generic_module_path, generic_object_path,
                        child_module_path, base_object_path,
                        generic_parameters_paths,
                        specialization_args, specializations_ast_nodes,
                        specializations_module_scope,
                        specializations_nodes_kinds,
                        specializations_references, modules_ast_nodes,
                        modules_definitions, modules_nodes_kinds,
                        modules_references, modules_submodules,
                        modules_superclasses
                )
                assert base_object_path not in modules_superclasses.get(
                        specializations_module_path, {}
                )
                specialization_table = dict(zip(generic_parameters_paths,
                                                specialization_args))
                specialize = _SpecializeGeneric(specialization_table).visit
                generic_bases = classes_bases[(generic_module_path,
                                               generic_object_path)]
                for generic_base in generic_bases:
                    base_base_module_path: _catalog.Path
                    base_base_object_path: _catalog.Path
                    if _is_generic_specialization(generic_base):
                        base_base_node = specialize(generic_base)
                        base_base_name = _serialization.to_identifier(
                                base_base_node
                        )
                        base_base_module_path, base_base_object_path = (
                            specializations_module_path, (base_base_name,)
                        )
                        if base_base_name not in specializations_module_scope:
                            generic_base_base_object_path = (
                                _conversion.to_path(generic_base.value)
                            )
                            (
                                generic_base_base_module_path,
                                generic_base_base_object_path
                            ) = _scoping.resolve_object_path(
                                    generic_module_path,
                                    generic_object_path[:-1],
                                    generic_base_base_object_path,
                                    modules_definitions, modules_references,
                                    modules_submodules, modules_superclasses
                            )
                            generic_base_base_parameters_paths = (
                                generics_parameters_paths[
                                    (generic_base_base_module_path,
                                     generic_base_base_object_path)
                                ]
                            )
                            base_base_specialization_args = _unpack_ast_node(
                                    _subscript_to_item(base_base_node)
                            )
                            _register_generic_specialization(
                                    generic_base_base_module_path,
                                    generic_base_base_object_path,
                                    child_module_path, base_base_object_path,
                                    generic_base_base_parameters_paths,
                                    base_base_specialization_args,
                                    specializations_ast_nodes,
                                    specializations_module_scope,
                                    specializations_nodes_kinds,
                                    specializations_references,
                                    modules_ast_nodes, modules_definitions,
                                    modules_nodes_kinds, modules_references,
                                    modules_submodules, modules_superclasses
                            )
                    else:
                        base_base_object_path = _conversion.to_path(
                                generic_base
                        )
                        base_base_module_path, base_base_object_path = (
                            _scoping.resolve_object_path(
                                    generic_module_path, (),
                                    base_base_object_path, modules_definitions,
                                    modules_references, modules_submodules,
                                    modules_superclasses
                            )
                        )
                    modules_superclasses.setdefault(
                            specializations_module_path, {}
                    ).setdefault(base_object_path, []).append(
                            (base_base_module_path, base_base_object_path)
                    )
            return specializations_module_path, base_object_path
        else:
            base_reference_path = _conversion.to_path(base)
            return _scoping.resolve_object_path(
                    child_module_path, child_object_path[:-1],
                    base_reference_path, modules_definitions,
                    modules_references, modules_submodules,
                    modules_superclasses
            )


    def _register_generic_specialization(
            generic_module_path: _catalog.Path,
            generic_object_path: _catalog.Path,
            specialization_module_path: _catalog.Path,
            specialization_object_path: _catalog.Path,
            generic_parameters_paths: _t.Sequence[_catalog.Path],
            specialization_args: _t.Sequence[_ast.expr],
            specializations_ast_nodes: _ModuleAstNodes,
            specializations_module_scope: _Scope,
            specializations_nodes_kinds: _ModuleAstNodesKinds,
            specializations_references: _ModuleReferences,
            modules_ast_nodes: _t.Dict[_catalog.Path, _ModuleAstNodes],
            modules_definitions: _t.Dict[_catalog.Path, _Scope],
            modules_nodes_kinds: _t.Dict[_catalog.Path, _ModuleAstNodesKinds],
            modules_references: _t.Dict[_catalog.Path, _ModuleReferences],
            modules_submodules: _t.Mapping[_catalog.Path, _ModuleSubmodules],
            modules_superclasses: _t.Mapping[
                _catalog.Path, _ModuleSuperclasses
            ],
            builtins_module_path: _catalog.Path
            = _catalog.module_path_from_module(_builtins)
    ) -> None:
        base_scope = specializations_module_scope
        for part in specialization_object_path:
            base_scope = base_scope.setdefault(part, {})
        generic_scope = modules_definitions[generic_module_path]
        for part in generic_object_path:
            generic_scope = generic_scope[part]
        base_scope.update(generic_scope)
        generic_module_ast_nodes = modules_ast_nodes[generic_module_path]
        specializations_nodes_kinds[
            specialization_object_path
        ] = _NodeKind.CLASS
        specialize = _SpecializeGeneric(
                dict(zip(generic_parameters_paths, specialization_args))
        ).visit
        args_names = (
                {arg.id
                 for arg in specialization_args
                 if _is_dependency_name(arg)}
                | {child.id
                   for ast_node in specialization_args
                   for child in _recursively_iterate_children(ast_node)
                   if _is_dependency_name(child)}
        )
        for name in generic_scope.keys():
            generic_field_path = (*generic_object_path, name)
            generic_ast_nodes = generic_module_ast_nodes[generic_field_path]
            specialization_field_path = (*specialization_object_path, name)
            specialization_ast_nodes = specializations_ast_nodes[
                specialization_field_path
            ] = [specialize(_deepcopy(ast_node))
                 for ast_node in generic_ast_nodes]
            specialization_definitions_names = set(
                    chain.from_iterable(
                            _conversion.to_names(ast_node)
                            for ast_node in specialization_ast_nodes
                    )
            )
            for dependency_name in (
                    {child.id
                     for ast_node in specialization_ast_nodes
                     for child in _recursively_iterate_children(ast_node)
                     if _is_dependency_name(child)}
                    - specialization_definitions_names
            ):
                dependency_path = (dependency_name,)
                dependency_module_path, dependency_object_path = (
                    _scoping.resolve_object_path(
                            (specialization_module_path
                             if dependency_name in args_names
                             else generic_module_path), (),
                            dependency_path, modules_definitions,
                            modules_references, modules_submodules,
                            modules_superclasses
                    )
                )
                if dependency_module_path != builtins_module_path:
                    specializations_references[dependency_path] = (
                        dependency_module_path, dependency_object_path
                    )
            specializations_nodes_kinds[specialization_field_path] = (
                modules_nodes_kinds[generic_module_path][generic_field_path]
            )


    if _execution.is_main_process():
        (
            definitions, nodes_kinds, raw_ast_nodes, references, submodules,
            superclasses
        ) = _execution.call_in_process(_parse_stubs_state,
                                       _stdlib_modules_paths)
        _caching.save(**{_DEFINITIONS_FIELD_NAME: definitions,
                         _NODES_KINDS_FIELD_NAME: nodes_kinds,
                         _RAW_AST_NODES_FIELD_NAME: raw_ast_nodes,
                         _REFERENCES_FIELD_NAME: references,
                         _SUBMODULES_FIELD_NAME: submodules,
                         _SUPERCLASSES_FIELD_NAME: superclasses,
                         _VERSION_FIELD_NAME: _version},
                      path=_CACHE_PATH)
    else:
        (
            definitions, nodes_kinds, raw_ast_nodes, references, submodules,
            superclasses
        ) = {}, {}, {}, {}, {}, {}



import ast
import sys
import typing as t
from pathlib import Path

from .conversion import RawAstNode

_AST_NAMESPACE = vars(ast).copy()
if sys.version_info >= (3, 8):
    del _AST_NAMESPACE[ast.Ellipsis.__qualname__]

if sys.version_info < (3, 8):
    def from_node(ast_node: ast.AST) -> ast.Module:
        return ast.Module([ast_node])
else:
    def from_node(ast_node: ast.AST) -> ast.Module:
        return ast.Module([ast_node], [])


def from_raw(raw: RawAstNode,
             *,
             namespace: t.Dict[str, t.Any] = _AST_NAMESPACE) -> ast.AST:
    result = eval(raw, namespace)
    assert isinstance(result, ast.AST), result
    return result


def from_source_path(path: Path) -> ast.Module:
    return ast.parse(path.read_text())



import ast as _ast
import builtins as _builtins
import inspect as _inspect
import sys as _sys
import types as _types
import typing as _t
from functools import (partial as _partial,
                       singledispatch as _singledispatch)
from itertools import zip_longest as _zip_longest

from typing_extensions import Self as _Self

from . import (catalog as _catalog,
               scoping as _scoping,
               stubs as _stubs)
from .arboreal import (construction as _construction,
                       conversion as _conversion)
from .arboreal.evaluation import (
    evaluate_expression_node as _evaluate_expression_node
)
from .arboreal.kind import NodeKind as _NodeKind
from .arboreal.utils import subscript_to_item as _subscript_to_item
from .models import (OptionalParameter as _OptionalParameter,
                     Parameter as _Parameter,
                     ParameterKind as _ParameterKind,
                     PlainSignature as _PlainSignature,
                     RequiredParameter as _RequiredParameter,
                     Signature as _Signature,
                     from_signatures as _from_signatures)
from .modules import supported_stdlib_qualified_paths as _qualified_paths
from .utils import decorate_if as _decorate_if


@_singledispatch
def from_callable(_callable: _t.Callable[..., _t.Any]) -> _Signature:
    raise TypeError(type(_callable))


@from_callable.register(_types.BuiltinFunctionType)
@_decorate_if(from_callable.register(_types.BuiltinMethodType),
              _sys.implementation.name != 'pypy')
def _(
        _callable: _t.Union[
            _types.BuiltinFunctionType, _types.BuiltinMethodType
        ]
) -> _Signature:
    try:
        return ((_from_callable(_callable)
                 if isinstance(_callable.__self__, type)
                 else (_from_callable(getattr(type(_callable.__self__),
                                              _callable.__name__))
                       .bind(_callable.__self__)))
                if (_callable.__self__ is not None
                    and not isinstance(_callable.__self__, _types.ModuleType))
                else _from_callable(_callable))
    except _SignatureNotFound:
        return _from_raw_signature(_inspect.signature(_callable))


@from_callable.register(_types.FunctionType)
def _(_callable: _types.FunctionType) -> _Signature:
    try:
        return _from_callable(_callable)
    except _SignatureNotFound:
        return _from_raw_signature(_inspect.signature(_callable))


@from_callable.register(_types.MethodType)
def _(_callable: _types.MethodType) -> _Signature:
    try:
        return (_from_callable(_callable)
                if isinstance(_callable.__self__, type)
                else (_from_callable(getattr(type(_callable.__self__),
                                             _callable.__name__))
                      .bind(_callable.__self__)))
    except _SignatureNotFound:
        return _from_raw_signature(_inspect.signature(_callable))


@_decorate_if(from_callable.register(_types.MethodWrapperType),
              _sys.implementation.name != 'pypy')
def _(_callable: _types.MethodWrapperType) -> _Signature:
    self = _callable.__self__
    assert not isinstance(self, type), _callable
    try:
        return (_from_callable(getattr(type(self), _callable.__name__))
                .bind(self))
    except _SignatureNotFound:
        return _from_raw_signature(_inspect.signature(_callable))


@_decorate_if(from_callable.register(_types.MethodDescriptorType),
              _sys.implementation.name != 'pypy')
@_decorate_if(from_callable.register(_types.WrapperDescriptorType),
              _sys.implementation.name != 'pypy')
def _(
        _callable: _t.Union[
            _types.MethodDescriptorType, _types.WrapperDescriptorType
        ]
) -> _Signature:
    cls = _callable.__objclass__
    assert isinstance(cls, type), _callable
    try:
        return _from_callable(_callable)
    except _SignatureNotFound:
        return _from_raw_signature(_inspect.signature(_callable))


@from_callable.register(type)
def _(_callable: _t.Callable[..., _t.Any]) -> _Signature:
    try:
        qualified_paths = resolve_qualified_paths(_callable)
        if not qualified_paths:
            raise _SignatureNotFound
        module_path, object_path = _resolve_builder_qualified_path(
                qualified_paths
        )
        ast_nodes = _load_ast_nodes(module_path, object_path)
        class_path, builder_name = object_path[:-1], object_path[-1]
        return _from_signatures(*[_from_statement_node(ast_node, _callable,
                                                       module_path, class_path)
                                  for ast_node in ast_nodes])
    except _SignatureNotFound:
        raw_signature = _inspect.signature(_callable)
        return _from_raw_signature(
                raw_signature.replace(return_annotation=_Self)
        )


@from_callable.register(_partial)
def _(_callable: '_partial[_t.Any]') -> _Signature:
    return from_callable(_callable.func).bind(*_callable.args,
                                              **_callable.keywords)


@_singledispatch
def _from_expression_node(ast_node: _ast.expr,
                          callable_: _t.Callable[..., _t.Any],
                          module_path: _catalog.Path,
                          parent_path: _catalog.Path) -> _Signature:
    raise TypeError(ast_node)


@_from_expression_node.register(_ast.Attribute)
@_from_expression_node.register(_ast.Name)
def _(ast_node: _t.Union[_ast.Attribute, _ast.Name],
      callable_: _t.Callable[..., _t.Any],
      module_path: _catalog.Path,
      parent_path: _catalog.Path,
      *,
      call_name: str = object.__call__.__name__) -> _Signature:
    object_path = _conversion.to_path(ast_node)
    module_path, object_path = _scoping.resolve_object_path(
            module_path, parent_path, object_path, _stubs.definitions,
            _stubs.references, _stubs.submodules, _stubs.superclasses
    )
    node_kind = _stubs.nodes_kinds[module_path][object_path]
    if node_kind is _NodeKind.CLASS:
        call_ast_nodes = _load_ast_nodes(module_path,
                                         (*object_path, call_name))
        call_signatures = [_from_statement_node(ast_node, callable_,
                                                module_path, ())
                           for ast_node in call_ast_nodes]
        return _from_signatures(*[signature.bind(callable_)
                                  for signature in call_signatures])
    else:
        annotation_nodes = [
            _construction.from_raw(raw)
            for raw in _stubs.raw_ast_nodes[module_path][object_path]
        ]
        if len(annotation_nodes) == 1:
            annotation_node, = annotation_nodes
            assert isinstance(annotation_node, _ast.stmt), (module_path,
                                                            object_path)
            return _from_statement_node(annotation_node, callable_,
                                        module_path, ())
    raise _SignatureNotFound


@_from_expression_node.register(_ast.Call)
def _(ast_node: _ast.Call,
      callable_: _t.Callable[..., _t.Any],
      module_path: _catalog.Path,
      parent_path: _catalog.Path,
      *,
      type_var_object_path: _catalog.Path
      = _catalog.path_from_string(_t.TypeVar.__qualname__),
      typing_module_path: _catalog.Path
      = _catalog.module_path_from_module(_t)) -> _Signature:
    callable_object_path = _conversion.to_path(ast_node.func)
    callable_module_path, callable_object_path = _scoping.resolve_object_path(
            module_path, parent_path, callable_object_path, _stubs.definitions,
            _stubs.references, _stubs.submodules, _stubs.superclasses
    )
    if (callable_module_path == typing_module_path
            and callable_object_path == type_var_object_path):
        maybe_bound_type_node = next((keyword.value
                                      for keyword in ast_node.keywords
                                      if keyword.arg == 'bound'),
                                     None)
        return (
            _from_signatures(
                    *[_from_expression_node(argument, callable_, module_path,
                                            parent_path)
                      for argument in ast_node.args[1:]]
            )
            if maybe_bound_type_node is None
            else _from_expression_node(maybe_bound_type_node, callable_,
                                       module_path, parent_path)
        )
    raise _SignatureNotFound


@_from_expression_node.register(_ast.Subscript)
def _(ast_node: _ast.Subscript,
      callable_: _t.Callable[..., _t.Any],
      module_path: _catalog.Path,
      parent_path: _catalog.Path,
      *,
      callable_object_path: _catalog.Path = ('Callable',),
      typing_module_path: _catalog.Path
      = _catalog.module_path_from_module(_t)) -> _Signature:
    value_path = _conversion.to_path(ast_node.value)
    value_module_path, value_object_path = _scoping.resolve_object_path(
            module_path, parent_path, value_path, _stubs.definitions,
            _stubs.references, _stubs.submodules, _stubs.superclasses
    )
    if (value_module_path == typing_module_path
            and value_object_path == callable_object_path):
        callable_arguments = _subscript_to_item(ast_node)
        assert isinstance(callable_arguments, _ast.Tuple)
        arguments_annotations, returns_annotation = callable_arguments.elts
        return (
            _PlainSignature(
                    *[
                        _RequiredParameter(
                                annotation=_evaluate_expression_node(
                                        annotation, module_path, parent_path,
                                        {}
                                ),
                                kind=_ParameterKind.POSITIONAL_ONLY,
                                name='_' + str(index),
                        )
                        for index, annotation in enumerate(
                                arguments_annotations.elts
                        )
                    ],
                    returns=_evaluate_expression_node(
                            returns_annotation, module_path, parent_path, {}
                    )
            )
            if isinstance(arguments_annotations, _ast.List)
            # unspecified parameters case
            else _PlainSignature(
                    _OptionalParameter(annotation=_t.Any,
                                       kind=_ParameterKind.VARIADIC_POSITIONAL,
                                       name='args'),
                    _OptionalParameter(annotation=_t.Any,
                                       kind=_ParameterKind.VARIADIC_KEYWORD,
                                       name='kwargs'),
                    returns=_evaluate_expression_node(
                            returns_annotation, module_path, parent_path, {}
                    )
            )
        )
    raise _SignatureNotFound


@_singledispatch
def _from_statement_node(ast_node: _ast.stmt,
                         callable_: _t.Callable[..., _t.Any],
                         module_path: _catalog.Path,
                         parent_path: _catalog.Path) -> _Signature:
    raise TypeError(ast_node)


@_from_statement_node.register(_ast.AnnAssign)
def _(ast_node: _ast.AnnAssign,
      callable_: _t.Callable[..., _t.Any],
      module_path: _catalog.Path,
      parent_path: _catalog.Path) -> _Signature:
    return _from_expression_node((ast_node.annotation
                                  if ast_node.value is None
                                  else ast_node.value), callable_, module_path,
                                 parent_path)


@_from_statement_node.register(_ast.Assign)
def _(ast_node: _ast.Assign,
      callable_: _t.Callable[..., _t.Any],
      module_path: _catalog.Path,
      parent_path: _catalog.Path) -> _Signature:
    return _from_expression_node(ast_node.value, callable_, module_path,
                                 parent_path)


@_from_statement_node.register(_ast.AsyncFunctionDef)
@_from_statement_node.register(_ast.FunctionDef)
def _(
        ast_node: _t.Union[_ast.AsyncFunctionDef, _ast.FunctionDef],
        callable_: _t.Callable[..., _t.Any],
        module_path: _catalog.Path,
        parent_path: _catalog.Path
) -> _Signature:
    parameters = _parameters_from(ast_node, callable_, module_path,
                                  parent_path)
    returns = _returns_annotation_from(ast_node, callable_, module_path,
                                       parent_path)
    return _PlainSignature(*parameters,
                           returns=returns)


def _parameters_from(
        ast_node: _t.Union[_ast.AsyncFunctionDef, _ast.FunctionDef],
        callable_: _t.Callable[..., _t.Any],
        module_path: _catalog.Path,
        parent_path: _catalog.Path
) -> _t.List[_Parameter]:
    signature_ast = ast_node.args
    result: _t.List[_Parameter] = list(filter(
            None,
            (*_to_positional_parameters(signature_ast, module_path,
                                        parent_path),
             _to_variadic_positional_parameter(signature_ast, module_path,
                                               parent_path),
             *_to_keyword_parameters(signature_ast, module_path, parent_path),
             _to_variadic_keyword_parameter(signature_ast, module_path,
                                            parent_path))
    ))
    if isinstance(callable_, type):
        del result[0]
    elif any(_is_classmethod(decorator_node, module_path, parent_path)
             for decorator_node in ast_node.decorator_list):
        result[0] = _RequiredParameter(annotation=_t.Type[_Self],
                                       kind=_ParameterKind.POSITIONAL_ONLY,
                                       name=result[0].name)
    elif _stubs.nodes_kinds[module_path].get(parent_path) is _NodeKind.CLASS:
        result[0] = _RequiredParameter(annotation=_Self,
                                       kind=_ParameterKind.POSITIONAL_ONLY,
                                       name=result[0].name)
    return result


def _returns_annotation_from(
        ast_node: _t.Union[_ast.AsyncFunctionDef, _ast.FunctionDef],
        callable_: _t.Callable[..., _t.Any],
        module_path: _catalog.Path,
        parent_path: _catalog.Path,
        *,
        initializer_name: str = object.__init__.__name__
) -> _t.Any:
    returns_node = ast_node.returns
    return (_Self
            if (isinstance(callable_, type)
                and ast_node.name == initializer_name)
            else (_t.Any
                  if returns_node is None
                  else _evaluate_expression_node(returns_node, module_path,
                                                 parent_path, {})))


def _is_classmethod(
        expression_node: _ast.expr,
        module_path: _catalog.Path,
        parent_path: _catalog.Path,
        *,
        classmethod_qualified_path: _catalog.QualifiedPath
        = (_catalog.module_path_from_module(_builtins),
           _catalog.path_from_string(classmethod.__qualname__)),
) -> bool:
    maybe_path = _conversion.to_maybe_path(expression_node)
    return (maybe_path is not None
            and _scoping.resolve_object_path(
                    module_path, parent_path, maybe_path, _stubs.definitions,
                    _stubs.references, _stubs.submodules, _stubs.superclasses
            ) == classmethod_qualified_path)


class _SignatureNotFound(Exception):
    pass


def _try_resolve_object_path(
        module_path: _catalog.Path, object_path: _catalog.Path
) -> _catalog.QualifiedPath:
    try:
        return _scoping.resolve_object_path(
                module_path, (), object_path, _stubs.definitions,
                _stubs.references, _stubs.submodules, _stubs.superclasses
        )
    except _scoping.ObjectNotFound:
        return (), ()


def _from_callable(value: _t.Callable[..., _t.Any]) -> _Signature:
    for module_path, object_path in _to_qualified_paths(value):
        ast_nodes = _load_ast_nodes(module_path, object_path)
        parent_path = object_path[:-1]
        try:
            signatures = [
                _from_statement_node(ast_node, value, module_path, parent_path)
                for ast_node in ast_nodes
            ]
        except _SignatureNotFound:
            continue
        else:
            return _from_signatures(*signatures)
    raise _SignatureNotFound


def _to_qualified_paths(
        value: _t.Callable[..., _t.Any]
) -> _t.Iterable[_catalog.QualifiedPath]:
    qualified_paths = resolve_qualified_paths(value)
    if qualified_paths:
        module_path, object_path = qualified_paths[0]
        if _stubs.nodes_kinds[module_path][object_path] is _NodeKind.CLASS:
            yield _resolve_builder_qualified_path(qualified_paths)
        else:
            yield from qualified_paths


def _resolve_builder_qualified_path(
        qualified_paths: _t.Sequence[_catalog.QualifiedPath],
        *,
        object_builder_qualified_path: _catalog.QualifiedPath
        = (_catalog.module_path_from_module(_builtins),
           _catalog.path_from_string(object.__qualname__)
           + _catalog.path_from_string(object.__new__.__name__))
) -> _catalog.QualifiedPath:
    candidates = {_to_class_builder_qualified_path(module_path, object_path)
                  for module_path, object_path in qualified_paths}
    try:
        (module_path, object_path), = candidates
    except ValueError:
        try:
            candidates.remove(object_builder_qualified_path)
            (module_path, object_path), = candidates
        except (KeyError, ValueError):
            raise _SignatureNotFound
    return module_path, object_path


def _load_ast_nodes(module_path: _catalog.Path,
                    object_path: _catalog.Path) -> _t.List[_ast.stmt]:
    try:
        raw_ast_nodes = _stubs.raw_ast_nodes[module_path][object_path]
    except KeyError:
        raise _SignatureNotFound
    else:
        assert raw_ast_nodes, (module_path, object_path)
        return [_statement_node_from_raw(raw_ast_node)
                for raw_ast_node in raw_ast_nodes]


def _statement_node_from_raw(raw: _conversion.RawAstNode) -> _ast.stmt:
    result = _construction.from_raw(raw)
    assert isinstance(result, _ast.stmt), raw
    return result


def resolve_qualified_paths(
        value: _t.Callable[..., _t.Any]
) -> _t.List[_catalog.QualifiedPath]:
    module_path, object_path = _catalog.qualified_path_from(value)
    try:
        candidates_paths = _qualified_paths[module_path][object_path]
    except KeyError:
        assert not module_path or object_path, value
        qualified_paths = ([(module_path, object_path)]
                           if module_path
                           else [])
    else:
        qualified_paths = [path
                           for path in candidates_paths
                           if _value_has_qualified_path(value, path)]
    return sorted({
        (module_path, object_path)
        for module_path, object_path in [
            _try_resolve_object_path(module_path, object_path)
            for module_path, object_path in qualified_paths
        ]
        if module_path and object_path
    })


def _to_class_builder_qualified_path(
        module_path: _catalog.Path,
        object_path: _catalog.Path,
        *,
        object_builder_qualified_path: _catalog.QualifiedPath
        = (_catalog.module_path_from_module(_builtins),
           _catalog.path_from_string(object.__qualname__)
           + _catalog.path_from_string(object.__new__.__name__)),
        constructor_name: str = object.__new__.__name__,
        initializer_name: str = object.__init__.__name__
) -> _catalog.QualifiedPath:
    for base_module_path, base_object_path in _to_mro(module_path,
                                                      object_path):
        base_module_annotations = _stubs.raw_ast_nodes[base_module_path]
        constructor_path = (*object_path, constructor_name)
        initializer_path = (*object_path, initializer_name)
        if initializer_path in base_module_annotations:
            return (base_module_path, initializer_path)
        elif constructor_path in base_module_annotations:
            return (base_module_path, constructor_path)
    return object_builder_qualified_path


def _to_mro(module_path: _catalog.Path,
            object_path: _catalog.Path) -> _t.Iterable[_catalog.QualifiedPath]:
    yield (module_path, object_path)
    try:
        bases = _stubs.superclasses[module_path][object_path]
    except KeyError:
        return
    else:
        for base_module_path, base_object_path in bases:
            yield from _to_mro(base_module_path, base_object_path)


def _value_has_qualified_path(value: _t.Any,
                              path: _catalog.QualifiedPath) -> bool:
    module_path, object_path = path
    module_name = _catalog.path_to_string(module_path)
    candidate = _sys.modules.get(module_name)
    if candidate is None:
        # undecidable, let's keep it
        return True
    for part in object_path:
        try:
            candidate = getattr(candidate, part)
        except AttributeError:
            return False
    return candidate is value


def _parameter_from_raw(raw: _inspect.Parameter) -> _Parameter:
    annotation, kind, name = (
        _t.Any if raw.annotation is _inspect._empty else raw.annotation,
        _ParameterKind(raw.kind), raw.name
    )
    return ((_OptionalParameter(annotation=annotation,
                                kind=kind,
                                name=name)
             if (kind is _ParameterKind.VARIADIC_POSITIONAL
                 or kind is _ParameterKind.VARIADIC_KEYWORD)
             else _RequiredParameter(annotation=annotation,
                                     kind=kind,
                                     name=name))
            if raw.default is _inspect._empty
            else _OptionalParameter(annotation=annotation,
                                    default=raw.default,
                                    kind=kind,
                                    name=name))


def _from_raw_signature(value: _inspect.Signature) -> _Signature:
    return _PlainSignature(
            *[_parameter_from_raw(raw) for raw in value.parameters.values()],
            returns=(_t.Any
                     if value.return_annotation is _inspect._empty
                     else value.return_annotation)
    )


def _parameter_from_ast_node(ast_node: _ast.arg,
                             default_ast: _t.Optional[_ast.expr],
                             module_path: _catalog.Path,
                             parent_path: _catalog.Path,
                             kind: _ParameterKind) -> _Parameter:
    annotation = (_t.Any
                  if ast_node.annotation is None
                  else _evaluate_expression_node(ast_node.annotation,
                                                 module_path, parent_path, {}))
    name = ast_node.arg
    if default_ast is not None:
        default = _evaluate_expression_node(default_ast, module_path,
                                            parent_path, {})
        return _OptionalParameter(annotation=annotation,
                                  **({}
                                     if default is Ellipsis
                                     else {'default': default}),
                                  kind=kind,
                                  name=name)
    elif (kind is _ParameterKind.VARIADIC_POSITIONAL
          or kind is _ParameterKind.VARIADIC_KEYWORD):
        return _OptionalParameter(annotation=annotation,
                                  kind=kind,
                                  name=name)
    else:
        return _RequiredParameter(annotation=annotation,
                                  kind=kind,
                                  name=name)


def _to_keyword_parameters(
        signature_ast: _ast.arguments,
        module_path: _catalog.Path,
        parent_path: _catalog.Path,
        *,
        kind: _ParameterKind = _ParameterKind.KEYWORD_ONLY
) -> _t.List[_Parameter]:
    return [_parameter_from_ast_node(parameter_ast, default_ast, module_path,
                                     parent_path, kind)
            for parameter_ast, default_ast in zip(signature_ast.kwonlyargs,
                                                  signature_ast.kw_defaults)]


def _to_positional_parameters(
        signature_ast: _ast.arguments,
        module_path: _catalog.Path,
        parent_path: _catalog.Path
) -> _t.List[_Parameter]:
    # double-reversing since parameters with default arguments go last
    parameters_with_defaults_ast: _t.List[
        _t.Tuple[_ast.arg, _t.Optional[_ast.expr]]
    ] = list(_zip_longest(reversed(signature_ast.args),
                          signature_ast.defaults))[::-1]
    kind = _ParameterKind.POSITIONAL_ONLY
    return [_parameter_from_ast_node(parameter_ast, default_ast, module_path,
                                     parent_path, kind)
            for parameter_ast, default_ast in parameters_with_defaults_ast]


def _to_variadic_keyword_parameter(
        signature_ast: _ast.arguments,
        module_path: _catalog.Path,
        parent_path: _catalog.Path,
        *,
        kind: _ParameterKind = _ParameterKind.VARIADIC_KEYWORD
) -> _t.Optional[_Parameter]:
    ast_node = signature_ast.kwarg
    return (None
            if ast_node is None
            else _parameter_from_ast_node(ast_node, None, module_path,
                                          parent_path, kind))


def _to_variadic_positional_parameter(
        signature_ast: _ast.arguments,
        module_path: _catalog.Path,
        parent_path: _catalog.Path,
        *,
        kind: _ParameterKind = _ParameterKind.VARIADIC_POSITIONAL
) -> _t.Optional[_Parameter]:
    ast_node = signature_ast.vararg
    return (None
            if ast_node is None
            else _parameter_from_ast_node(ast_node, None, module_path,
                                          parent_path, kind))



import traceback
import typing as t
from functools import singledispatch


def format_exception(value: BaseException) -> str:
    return ''.join(traceback.format_exception(type(value), value,
                                              value.__traceback__))


@singledispatch
def repr_from(value: t.Any, indent: int, depth: int) -> str:
    return repr(value)


@repr_from.register(dict)
def _(value: t.Dict[t.Any, t.Any], indent: int, depth: int) -> str:
    return (
        ('{\n'
         + ',\n'.join(sorted([
                    indent * ' ' * (depth + 1)
                    + repr_from(key, indent, depth + 1)
                    + ': '
                    + repr_from(sub_value, indent, depth + 1)
                    for key, sub_value in value.items()

                ]))
         + '\n' + indent * ' ' * depth + '}')
        if value
        else repr(value)
    )


@repr_from.register(list)
def _(value: t.List[t.Any], indent: int, depth: int) -> str:
    return (('[\n'
             + ',\n'.join([indent * ' ' * (depth + 1)
                           + repr_from(sub_value, indent, depth + 1)
                           for sub_value in value])
             + '\n' + indent * ' ' * depth + ']')
            if value
            else repr(value))


@repr_from.register(tuple)
def _(value: t.Tuple[t.Any, ...], indent: int, depth: int) -> str:
    if len(value) > 1:
        return ('(\n'
                + ',\n'.join([indent * ' ' * (depth + 1)
                              + repr_from(sub_value, indent, depth + 1)
                              for sub_value in value])
                + '\n' + indent * ' ' * depth + ')')
    elif value:
        return f'({repr_from(value[0], indent, depth)},)'
    else:
        return repr(value)


@repr_from.register(set)
def _(value: t.Set[t.Any], indent: int, depth: int) -> str:
    return (('{\n'
             + ',\n'.join([indent * ' ' * (depth + 1)
                           + repr_from(sub_value, indent, depth + 1)
                           for sub_value in value])
             + '\n' + indent * ' ' * depth + '}')
            if value
            else repr(value))



import inspect
import multiprocessing
import sys
import threading
import typing as t
from multiprocessing.queues import SimpleQueue


def is_main_process() -> bool:
    return multiprocessing.current_process().name == 'MainProcess'


def call_in_process(function: t.Callable[..., t.Any],
                    *args: t.Any,
                    **kwargs: t.Any) -> t.Any:
    with threading.Lock():
        caller_frame = inspect.stack()[1].frame
        caller_module = inspect.getmodule(caller_frame)
        assert caller_module is not None
        main_module = sys.modules.get('__main__')
        try:
            sys.modules['__main__'] = caller_module
            context = multiprocessing.get_context()
            queue = context.SimpleQueue()
            process = context.Process(
                    target=_put_result_in_queue,
                    name=(f'{call_in_process.__qualname__}_'
                          f'{function.__qualname__}'),
                    args=(queue, function, *args),
                    kwargs=kwargs
            )
            process.start()
            result, error = queue.get()
            process.join()
            if error is not None:
                raise error
            return result
        finally:
            if main_module is None:
                del sys.modules['__main__']
            else:
                sys.modules['__main__'] = main_module


def _put_result_in_queue(queue: 'SimpleQueue[t.Any]',
                         function: t.Callable[..., t.Any],
                         *args: t.Any,
                         **kwargs: t.Any) -> None:
    try:
        result = function(*args, **kwargs)
    except Exception as error:
        queue.put((None, error))
        raise error
    else:
        queue.put((result, None))



import sys as _sys
import typing as _t
from pathlib import Path as _Path

import mypy as _mypy
from mypy.version import __version__ as _mypy_version

from paradigm import __version__ as _version
from . import (caching as _caching,
               index as _index)

_CACHE_PATH = _Path(__file__).with_name(
        '_' + _mypy.__name__ + '_' + _mypy_version.replace('.', '_')
        + '_' + _sys.platform
        + '_' + _sys.implementation.name
        + '_' + '_'.join(map(str, _sys.version_info))
        + '_' + _Path(__file__).with_suffix(_caching.FILE_SUFFIX).name
)
_STDLIB_QUALIFIED_PATHS_FIELD_NAME = 'stdlib_qualified_paths'
_VERSION_FIELD_NAME = 'version'

supported_stdlib_qualified_paths: _index.QualifiedPaths
try:
    supported_stdlib_qualified_paths, _cached_version = _caching.load(
            _STDLIB_QUALIFIED_PATHS_FIELD_NAME, _VERSION_FIELD_NAME,
            path=_CACHE_PATH
    )
except Exception:
    _reload_cache = True
else:
    _reload_cache = _cached_version != _version
if _reload_cache:
    from . import execution as _execution

    if _execution.is_main_process():
        from . import (catalog as _catalog,
                       scoping as _scoping,
                       stubs as _stubs)
        from .discovery import (
            supported_stdlib_modules_paths as _supported_stdlib_modules_paths
        )


        def _to_supported_qualified_paths(
                qualified_paths: _index.QualifiedPaths,
                definitions: _t.Mapping[_catalog.Path, _scoping.Scope],
                references: _t.Mapping[
                    _catalog.Path, _scoping.ModuleReferences
                ],
                submodules: _t.Mapping[
                    _catalog.Path, _scoping.ModuleSubmodules
                ],
                superclasses: _t.Mapping[
                    _catalog.Path, _scoping.ModuleSuperclasses
                ]
        ) -> _index.QualifiedPaths:
            result = {}
            for (
                    module_path, module_qualified_paths
            ) in qualified_paths.items():
                supported_module_qualified_paths = {}
                for (
                        object_path, object_qualified_paths
                ) in module_qualified_paths.items():
                    supported_object_qualified_paths = [
                        (located_module_path, located_object_path)
                        for located_module_path, located_object_path
                        in object_qualified_paths
                        if _scoping.contains_object_path(
                                located_module_path, (), located_object_path,
                                definitions, references, submodules,
                                superclasses
                        )
                    ]
                    if supported_object_qualified_paths:
                        supported_module_qualified_paths[object_path] = (
                            supported_object_qualified_paths
                        )
                if supported_module_qualified_paths:
                    result[module_path] = supported_module_qualified_paths
            return result


        supported_stdlib_qualified_paths = _to_supported_qualified_paths(
                _execution.call_in_process(_index.from_modules,
                                           _supported_stdlib_modules_paths),
                _stubs.definitions, _stubs.references, _stubs.submodules,
                _stubs.superclasses
        )
        _caching.save(
                **{
                    _STDLIB_QUALIFIED_PATHS_FIELD_NAME:
                        supported_stdlib_qualified_paths,
                    _VERSION_FIELD_NAME: _version
                },
                path=_CACHE_PATH
        )
    else:
        supported_stdlib_qualified_paths = {}



from enum import (IntEnum,
                  auto)


class NodeKind(IntEnum):
    ANNOTATED_ASSIGNMENT = auto()
    ASSIGNMENT = auto()
    ASYNC_FUNCTION = auto()
    CLASS = auto()
    FUNCTION = auto()

    def __repr__(self) -> str:
        return repr(self.value)
